"""This module provides classes for batch normalisation of gait data in a trial."""

from abc import ABC, abstractmethod

import gaitalytics.model as model


class BaseNormaliser(ABC):
    """Base class for normalisers.

    This class provides a common interface for normalising data.
    """

    @abstractmethod
    def normalise(
        self, trial: model.Trial | model.TrialCycles
    ) -> model.Trial | model.TrialCycles:
        """Normalises the input data.

        Args:
            trial: The trial to be normalised.

        Returns:
            A new trial containing the normalised data.
        """
        raise NotImplementedError


class LinearTimeNormaliser(BaseNormaliser):
    """A class for normalising data based on time.

    This class provides a method to normalise the data based on time.
    """

    def __init__(self, n_frames: int = 100):
        """Initializes a new instance of the LinearTimeNormaliser class.

        Args:
            n_frames: The number of frames to time-normalise the data to.
        """
        self.n_frames: int = n_frames

    def normalise(
        self, trial: model.Trial | model.TrialCycles
    ) -> model.Trial | model.TrialCycles:
        """Normalises the data based on time.

        Args:
            trial: The trial to be normalised.

        Returns:
            A new trial containing the time-normalised data.
        """
        if type(trial) is model.TrialCycles:
            trial = self._normalise_cycle(trial)
        elif type(trial) is model.Trial:
            trial = self._normalise_trial(trial)
        return trial

    def _normalise_trial(self, trial: model.Trial) -> model.Trial:
        """Normalises the data of a Trial based on time.

        Args:
            trial: The trial to be normalised.

        Returns: A new trial containing the time-normalised data.

        """
        new_trial = model.Trial()
        for data_category in trial.get_all_data():
            data = trial.get_data(data_category)
            norm_data = data.meca.time_normalize(n_frames=self.n_frames, norm_time=True)

            new_trial.add_data(data_category, norm_data)
        return new_trial

    def _normalise_cycle(self, trial: model.TrialCycles) -> model.TrialCycles:
        """Normalises the data of a segmented Trial based on time.

        Args:
            trial: The trial to be normalised.

        Returns: A new trial containing the time-normalised data.
        """
        norm_cycles = model.TrialCycles()
        for context, cycles in trial.get_all_cycles().items():
            for cycle_id, cycle in cycles.items():
                new_cycle = self._normalise_trial(cycle)
                norm_cycles.add_cycle(context, cycle_id, new_cycle)

        return norm_cycles



from enum import Enum
from pathlib import Path

from yaml import safe_load


class MappedMarkers(Enum):
    """This defines the markers that are mapped in the configuration file.

    Attributes:
        L_HEEL (str): Left heel marker
        R_HEEL (str): Right heel marker
        L_TOE (str): Left toe marker
        R_TOE (str): Right toe marker
        L_ANKLE (str): Left ankle marker
        R_ANKLE (str): Right ankle marker
        L_TOE_2 (str): Left toe 2 marker
        R_TOE_2 (str): Right toe 2 marker
        L_ANT_HIP (str): Left anterior hip marker
        R_ANT_HIP (str): Right anterior hip marker
        L_POST_HIP (str): Left posterior hip marker
        R_POST_HIP (str): Right posterior hip marker
        SACRUM (str): Sacrum marker
        XCOM (str): Extrapolated center of mass marker

    """

    # Foot
    L_HEEL = "l_heel"
    R_HEEL = "r_heel"
    L_TOE = "l_toe"
    R_TOE = "r_toe"
    L_ANKLE = "l_lat_malleoli"
    R_ANKLE = "r_lat_malleoli"
    # Additional toe markers
    L_TOE_2 = "l_toe_2"
    R_TOE_2 = "r_toe_2"

    # Hip
    L_ANT_HIP = "l_ant_hip"
    R_POST_HIP = "r_post_hip"
    L_POST_HIP = "l_post_hip"
    R_ANT_HIP = "r_ant_hip"
    SACRUM = "sacrum"

    # Extrapolated center of mass marker for margin of stability
    XCOM = "xcom"


class MappingConfigs:
    """A class for reading the mapping configuration file.

    This class provides methods to read the mapping configuration file and get
    the markers and analogs for analysis.
    """

    _SEC_ANALYSIS: str = "analysis"
    _SEC_MARKERS_ANALYSIS: str = "markers"
    _SEC_ANALOGS_ANALYSIS: str = "analogs"

    _SEC_MAPPING: str = "mapping"
    _SEC_MARKERS_MAPPING: str = "markers"
    _SEC_ANALOGS_MAPPING: str = "analogs"

    def __init__(self, config_path: Path):
        """Initializes a new instance of the MappingConfigs class.

        Reads the yaml file into memory.

        Args:
            config_path: The path to the configuration file.
        """
        self._configs: dict[str, dict] = {}
        with open(config_path) as stream:
            self._configs = safe_load(stream)

    def get_markers_analysis(self) -> list[str]:
        """Gets the markers for analysis.

        Returns:
            A list of marker names to be used for analysis
            if present in the config file, otherwise an empty list.

        Raises:
            ValueError: If the analysis section is missing in the config file.
        """
        self._check_analysis_section()
        if self._SEC_MARKERS_ANALYSIS not in self._configs[self._SEC_ANALYSIS]:
            return []
        return self._configs[self._SEC_ANALYSIS][self._SEC_MARKERS_ANALYSIS]

    def get_analogs_analysis(self) -> list[str]:
        """Gets the analogs for analysis.

        Returns:
            A list of analog names to be used for analysis
            if present in the config file, otherwise an empty list.

        Raises:
            ValueError: If the analysis section is missing in the config file.
        """
        self._check_analysis_section()
        if self._SEC_ANALOGS_ANALYSIS not in self._configs[self._SEC_ANALYSIS]:
            return []
        return self._configs[self._SEC_ANALYSIS][self._SEC_ANALOGS_ANALYSIS]

    def _check_analysis_section(self):
        """Checks if the analysis section is present in the config file.

        Raises:
            ValueError: If the analysis section is missing in the config file.
        """
        if self._configs is None or self._SEC_ANALYSIS not in self._configs:
            raise ValueError("Analysis section is missing in the config file.")

    def get_marker_mapping(self, marker: MappedMarkers) -> str:
        """Gets the mapping of markers.

        Args:
            marker: The marker to get the mapping for.

        Returns:
            The mapped marker name if present in the config file.

        Raises:
            ValueError: If sections in the mapping are missing in the config file.
        """
        self._check_marker_mapping()

        return self._configs[self._SEC_MAPPING][self._SEC_MARKERS_MAPPING][marker.value]

    def _check_marker_mapping(self):
        """Checks if the marker mapping section is present in the config file.

        Raises:
            ValueError: If the mapping section is missing in the config file.
        """
        if self._configs is None or self._SEC_MAPPING not in self._configs:
            raise ValueError("Mapping section is missing in the config file.")
        elif self._SEC_MARKERS_MAPPING not in self._configs[self._SEC_MAPPING]:
            raise ValueError("Marker mapping section is missing in the config file.")



"""This module provides classes for reading biomechanical file-types."""

import math
from abc import abstractmethod, ABC
from pathlib import Path

import ezc3d
import numpy as np
import pandas as pd
import pyomeca
import xarray as xr

import gaitalytics.mapping as mapping
import gaitalytics.model as model

_MAX_EVENTS_PER_SECTION = 255


# Input Section
class _BaseFileHandler(ABC):
    """Base class for file handler.

    This class provides a common interface for file handlers.

    Attributes:
        file_path: The path to the  file.
    """

    def __init__(self, file_path: Path):
        """Initialize a new instance of the _BaseFileHandler class.

        Args:
            file_path: The path to the input file.
        """
        self.file_path = file_path


class _EventFileWriter(_BaseFileHandler, ABC):
    @abstractmethod
    def write_events(self, events: pd.DataFrame, file_path: Path | None = None):
        """Write the events to the output file.

        Args:
            events: The events to write to the output file.
            file_path: The path to the output file if deviating from the input file.
        """
        raise NotImplementedError


class C3dEventFileWriter(_EventFileWriter):
    """A class for handling C3D files in an easy and convenient way."""

    _CONTEXT_SECTION = "CONTEXTS"
    _ICON_SECTION = "ICON_IDS"
    _LABEL_SECTION = "LABELS"
    _TIME_SECTION = "TIMES"

    def write_events(self, events: pd.DataFrame, file_path: Path | None = None):
        """Write the events to the output file.

        Args:
            events: The events to write to the output file.
            file_path: The path to the output file if deviating from the input file.
        """
        c3d = ezc3d.c3d(str(self.file_path))
        n_sections = math.ceil(len(events) / _MAX_EVENTS_PER_SECTION)
        for i in range(n_sections):
            start = i * _MAX_EVENTS_PER_SECTION
            end = (i + 1) * _MAX_EVENTS_PER_SECTION
            if end > len(events):
                end = len(events)

            context_label = self._CONTEXT_SECTION
            icon_label = self._ICON_SECTION
            label_label = self._LABEL_SECTION
            time_label = self._TIME_SECTION

            if i > 0:
                context_label += f"{i + 1}"
                icon_label += f"{i + 1}"
                label_label += f"{i + 1}"
                time_label += f"{i + 1}"
            subset_events = events.iloc[start:end]
            c3d.add_parameter("EVENT", context_label, subset_events["context"].tolist())
            c3d.add_parameter("EVENT", icon_label, subset_events["icon_id"].tolist())
            c3d.add_parameter("EVENT", label_label, subset_events["label"].tolist())

            # times = [[time // 60, time % 60] for time in subset_events["time"].tolist()]
            raw_times = subset_events["time"].to_numpy()
            minutes = raw_times // 60
            seconds = raw_times % 60

            times = np.vstack((minutes, seconds))

            c3d.add_parameter("EVENT", time_label, times)
        c3d.add_parameter("EVENT", "USED", len(events))
        path = file_path if file_path else self.file_path
        c3d.write(str(path))


class _EventInputFileReader(_BaseFileHandler, ABC):
    """Abstract base class for reading event input files.

    This class defines the interface for reading event input files
    in gait analysis.
    Subclasses must implement the abstract methods to provide specific
    implementations.

    """

    COLUMN_TIME = "time"
    COLUMN_LABEL = "label"
    COLUMN_CONTEXT = "context"
    COLUMN_ICON = "icon_id"

    @abstractmethod
    def get_events(self) -> pd.DataFrame:
        """Get the events from the input file sorted by time.

        Returns:
            A DataFrame containing the events.
        """
        raise NotImplementedError


class C3dEventInputFileReader(_EventInputFileReader):
    """A class for handling C3D files in an easy and convenient way.

    Implements the EventInputFileReader interface to read events from C3D files.
    """

    def __init__(self, file_path: Path):
        """Initializes a new instance of the EzC3dFileHandler class.

        Args:
            file_path: The path to the C3D file.

        """
        self._c3d = ezc3d.c3d(str(file_path))
        super().__init__(file_path)

    def get_events(self) -> pd.DataFrame | None:
        """Gets the events from the input file sorted by time.

        Returns:
            pd.DataFrame A DataFrame containing the events.
            None if there are no events in the C3D file.


        """
        labels = self._get_event_labels()
        times = self._get_event_times()
        contexts = self._get_event_contexts()
        icons = self._get_event_icons()

        # Check if there are any events in the C3D file
        if not (labels and times and contexts and icons):
            return None

        table = pd.DataFrame(
            {
                self.COLUMN_TIME: times,
                self.COLUMN_LABEL: labels,
                self.COLUMN_CONTEXT: contexts,
                self.COLUMN_ICON: icons,
            }
        )
        table = table.sort_values(by=self.COLUMN_TIME, ascending=True).reset_index(
            drop=True
        )
        return table

    def _get_event_labels(self) -> list[str]:
        """Gets the labels of the events in the C3D file.

        Returns:
            The labels of the events.

        """
        section_base = "LABELS"
        labels = self._concat_sections(section_base)
        return labels

    def _get_event_times(self) -> list:
        """Returns the event times.

        Returns:
            An array containing the event times.
        """
        section_base = "TIMES"
        return self._concat_sections(section_base)

    def _get_event_contexts(self) -> list[str]:
        """Gets the contexts of the events in the C3D file.

        Returns:
            The contexts of the events.

        """
        section_base = "CONTEXTS"
        return self._concat_sections(section_base)

    def _get_event_icons(self) -> list:
        """Gets the icons of the events in the C3D file.

        Returns:
            The icons of the events.
        """
        section_base = "ICON_IDS"
        return self._concat_sections(section_base)

    def _get_sections(self, section_base):
        """Gets the sections of the specified type in the C3D file.

        Args:
            The base name of the sections to get.

        Returns:
            A list containing the sections of the specified type.
        """
        sections = []
        for section in self._c3d["parameters"]["EVENT"].keys():
            if section.startswith(section_base):
                sections.append(section)
        return sections

    def _concat_sections(self, section_base: str) -> list:
        """Concatenates the values of the specified sections in the C3D file.

        Args:
            section_base: The base name of the sections to concatenate.

        Returns:
            A list containing the concatenated values of the sections.
        """
        values: list = []
        for section in self._get_sections(section_base):
            current_values = self._c3d["parameters"]["EVENT"][section]["value"]
            if len(current_values) == 2:
                # convert TIMES: c3d specifics values[0] as
                # minutes and values[1] as seconds
                current_values = current_values[1] + (current_values[0] * 60)
                current_values = np.round(current_values, 3)
            if type(current_values) is np.ndarray:
                current_values = current_values.tolist()
            values += current_values
        return values


class _PyomecaInputFileReader(_BaseFileHandler, ABC):
    """Base class for handling input files using pyomeca.

    This class provides a common interface for reading input files with pyomeca.
    """

    def __init__(
        self, file_path: Path, pyomeca_class: type[pyomeca.Markers | pyomeca.Analogs]
    ):
        """Initializes a new instance of the MarkersInputFileReader class.

        Determines the file format and uses the appropriate pyomeca class
        to read the data. Further it converts the data to absolute time if needed.

        Args:
            file_path: The path to the marker data file.
            pyomeca_class:
                The pyomeca class to use for reading the data.

        """
        file_ext = file_path.suffix
        if file_ext == ".c3d" and (
            pyomeca_class == pyomeca.Analogs or pyomeca_class == pyomeca.Markers
        ):
            data = pyomeca_class.from_c3d(file_path)
        elif file_ext == ".trc" and pyomeca_class == pyomeca.Markers:
            raise NotImplementedError("TRC file format is not supported for markers")
        elif file_ext == ".mot" and pyomeca_class == pyomeca.Analogs:
            data = pyomeca_class.from_mot(
                file_path, pandas_kwargs={"sep": "\t", "index_col": False}
            )
        elif file_ext == ".sto" and pyomeca_class == pyomeca.Analogs:
            raise NotImplementedError("STO file format is not supported for analogs")
        else:
            raise ValueError(
                f"Unsupported file extension: {file_ext} for class {pyomeca_class}"
            )

        if "first_frame" in data.attrs and "rate" in data.attrs:
            first_frame = data.attrs["first_frame"]
            frame_rate = data.attrs["rate"]
            data = self._to_absolute_time(data, first_frame, frame_rate)

        self._data = data
        super().__init__(file_path)

    @staticmethod
    def _to_absolute_time(
        data: xr.DataArray, first_frame: int, rate: float
    ) -> xr.DataArray:
        """Converts the time to absolute time.

        Args:
            data: The data to convert.
            first_frame: The first frame of the data.
            rate: The rate of the data.
        """
        first_frame = first_frame
        data.coords["time"] = data.coords["time"] + (first_frame * 1 / rate)
        return data


class MarkersInputFileReader(_PyomecaInputFileReader):
    """A class for handling marker data in an easy and convenient way.

    Uses the pyomeca.Markers class to read marker data from a file.
    """

    def __init__(self, file_path: Path):
        """Initializes a new instance of the MarkersInputFileReader class.

        Args:
            file_path: The path to the marker data file.

        """
        super().__init__(file_path, pyomeca.Markers)
        self.data = self._data.drop_sel(axis="ones")

    def get_markers(self) -> xr.DataArray:
        """Gets the markers from the input file.

        Returns:
            An xarray DataArray containing the markers.
        """
        return self.data


class AnalogsInputFileReader(_PyomecaInputFileReader):
    """A class for handling analog data in an easy and convenient way.

    Uses the pyomeca.Analogs class to read analog data from a file.
    """

    def __init__(self, file_path: Path):
        """Initializes a new instance of the AnalogsInputFileReader class.

        Args:
            file_path: The path to the analog data file.

        """
        super().__init__(file_path, pyomeca.Analogs)

    def get_analogs(self) -> xr.DataArray:
        """Gets the analog data from the input file.

        Returns:
            An xarray DataArray containing the analog data.
        """
        return self._data


class AnalysisInputReader(_PyomecaInputFileReader):
    """Read out data from modelled data form different input format."""

    def __init__(self, file_path: Path, configs: mapping.MappingConfigs):
        """Initializes a new instance of the AnalysisInputReader class.

        Args:
            file_path: The path to the input file.
            configs: The mapping configurations.
        """
        extension = file_path.suffix
        pyomeca_class: type[pyomeca.Markers | pyomeca.Analogs]
        if extension == ".c3d":
            pyomeca_class = pyomeca.Markers
        elif extension == ".mot":
            pyomeca_class = pyomeca.Analogs
        elif extension == ".sto":
            raise NotImplementedError("STO file format is not supported for analogs")
        else:
            raise ValueError(f"Unsupported file extension: {extension}")
        super().__init__(file_path, pyomeca_class)
        self.configs = configs

        if pyomeca_class == pyomeca.Markers:
            self._data = self._data.drop_sel(axis="ones")
            self._filter_markers()
            self._flatten_array()
        else:
            self._filter_analogs()

    def _filter_analogs(self):
        """Filter the analogs based on the mapping configurations."""
        labels = self.configs.get_analogs_analysis()
        if labels:
            self._data = self._data.sel(channel=labels)

    def _filter_markers(self):
        """Filter the markers based on the mapping configurations."""
        labels = self.configs.get_markers_analysis()
        if labels:
            self._data = self._data.sel(channel=labels)

    def _flatten_array(self):
        """Flatten the markers array to a 2D array.

        The markers xarray is reshaped to a 2D array with the shape to generalize format
        from c3d and sto files.
        """
        np_data = self._data.to_numpy()
        rs_data = np_data.reshape(
            (np_data.shape[1] * np_data.shape[0], np_data.shape[2]), order="F"
        )

        time = self._data.coords["time"].values
        new_labels = self._create_labels()
        new_format = xr.DataArray(
            rs_data,
            coords={"channel": new_labels, "time": time},
            dims=["channel", "time"],
        )
        self._copy_attrs(new_format)
        self._data = new_format

    def _create_labels(self) -> list[str]:
        """Create new labels for the flattened array.

        Returns:
            List of labels including axis in the label name.
        """
        labels = self._data.coords["channel"].values
        axis = self._data.coords["axis"].values
        new_labels = []
        for label in labels:
            for axe in axis:
                new_labels.append(f"{label}_{axe}")
        return new_labels

    def _copy_attrs(self, new_data: xr.DataArray):
        """Copy the attributes from the old data to the new data.

        Args:
            new_data: The new data array.
        """
        new_data.attrs = self._data.attrs
        new_data.name = self._data.name

    def get_analysis(self) -> xr.DataArray:
        """Gets the analysis data from the input file.

        Returns:
            An xarray DataArray containing the analysis data.
        """
        return self._data


class _TrialExporter(_BaseFileHandler):
    @abstractmethod
    def export_trial(self, trial: model.Trial | model.TrialCycles):
        """Export the trial to the output folder.

        Args:
            trial: The trial to export.
            folder_path: The path to the output folder.
        """
        raise NotImplementedError


class NetCDFTrialExporter(_TrialExporter):
    """A class for exporting trial data to NetCDF files."""

    def export_trial(self, trial: model.Trial | model.TrialCycles):
        """Export the trial to the output folder.

        Args:
            trial: The trial to export.
            folder_path: The path to the output folder.
        """
        if not self.file_path.exists():
            self.file_path.mkdir(parents=True)

        if isinstance(trial, model.Trial):
            self._export_trial(trial, self.file_path)
        elif isinstance(trial, model.TrialCycles):
            self._export_trial_cycles(trial, self.file_path)

    @staticmethod
    def _export_trial(trial: model.Trial, folder_path: Path):
        """Export the trial to the output folder.

        Args:
            trial: The trial to export.
            folder_path: The path to the output folder.
        """

        for category, data in trial.get_all_data().items():
            file_path = folder_path / f"{category.value}.nc"
            data.to_netcdf(file_path, mode="w", engine="h5netcdf")
        if trial.events is not None:
            file_path = folder_path / "events.nc"
            trial.events.to_xarray().to_netcdf(file_path, mode="w", engine="h5netcdf")

    @staticmethod
    def _export_trial_cycles(trial: model.TrialCycles, folder_path: Path):
        """Export the trial cycles to the output folder.

        Args:
            trial: The trial cycles to export.
            folder_path: The path to the output folder.
        """
        for category in model.DataCategory:
            context_structs = {}
            category_exists = True
            for context, cont_segments in trial.get_all_cycles().items():
                cycle_struct: dict[int, xr.DataArray] = {}
                for cycle_id, cycle in cont_segments.items():
                    try:
                        cycle_data = cycle.get_data(category)
                        cycle_struct[cycle_id] = cycle_data
                    except KeyError:
                        category_exists = False
                        break
                if not category_exists:
                    break
                cycle_data = xr.Dataset(cycle_struct).to_dataarray("cycle")
                context_structs[context] = cycle_data

            if category_exists:
                file_path = folder_path / f"{category.value}.nc"
                full_data = xr.Dataset(context_structs).to_dataarray("context")
                full_data.to_netcdf(file_path, mode="w", engine="h5netcdf")



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
from importlib.metadata import version
release = version("gaitalytics")
#version = ".".join(release.split(".")[:2])
version = release

project = 'gaitalytics'
copyright = '2024, André Böni'
author = 'André Böni'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ['sphinx.ext.autodoc',
              'sphinx_codeautolink',
              'sphinxcontrib.bibtex',
              'sphinx.ext.napoleon',
              'sphinxcontrib.mermaid',]

bibtex_bibfiles = ['_static/Gaitalytics.bib']

bibtex_encoding = 'utf-8'
# bibtex_reference_style = "author_year"
# bibtex_default_style = "plain"

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']

html_theme_options = {
    'collapse_navigation': True,
    'sticky_navigation': True,
    'navigation_depth': 4,
    'includehidden': True,
    'titles_only': False,
    'prev_next_buttons_location': 'bottom',
}

html_favicon = '_static/favicon.png'


napoleon_google_docstring = True
napoleon_include_init_with_doc = True




"""This module contains classes for segmenting the trial data with different methods.

The module provides classes for segmenting the trial data based on
gait events as well as a base class to implement additional methods.
"""

from abc import ABC, abstractmethod

import numpy as np
import pandas as pd
import xarray as xr

import gaitalytics.events as ga_events
import gaitalytics.io as io
import gaitalytics.model as model
import gaitalytics.utils.math as ga_math


class _BaseSegmentation(ABC):
    @abstractmethod
    def segment(self, trial: model.Trial) -> model.TrialCycles:
        """Segments the trial data based on the segmentation method.

        Args:
            trial: The trial to be segmented.
        """
        raise NotImplementedError


class GaitEventsSegmentation(_BaseSegmentation):
    """A class for segmenting the trial data based on gait events.

    This class provides a method to segment the trial data based on gait events.
    It splits the trial data based on the event label and context.
    """

    def __init__(self, event_label: str = ga_events.FOOT_STRIKE):
        """Initializes a new instance of the GaitEventsSegmentation class.

        Args:
            event_label: The label of the event to be used for segmentation.
        """
        self.event_label = event_label

    def segment(self, trial: model.Trial) -> model.TrialCycles:
        """Segments the trial data based on gait events and contexts.

        Args:
            trial: The trial to be segmented.

        Returns:
            A new trial containing the all the cycles.

        Raises:
            ValueError: If the trial does not have events.
        """
        events = trial.events
        if events is None:
            raise ValueError("Trial does not have events.")

        events_times = self._get_times_of_events(events)

        trial_cycles = model.TrialCycles()
        for context, times in events_times.items():
            for cycle_id in range(len(times) - 1):
                start_time = times[cycle_id]
                end_time = times[cycle_id + 1]
                trial_cycles.add_cycle(
                    context,
                    cycle_id,
                    self._get_segment(trial, start_time, end_time, cycle_id, context),
                )

        return trial_cycles

    def _get_times_of_events(self, events: pd.DataFrame) -> dict[str, list]:
        """Gets the times of the events in the trial.

        This method splits the trial data based on the event label and context.

        Args:
            events: The events in the trial.

        Returns:
            A dictionary containing the contexts
            as keys and the event times as values.
        """
        splits = {}
        interesting_events = events[
            events[io._EventInputFileReader.COLUMN_LABEL] == self.event_label
        ]
        contexts = events[io._EventInputFileReader.COLUMN_CONTEXT].unique()
        for context in contexts:
            context_events = interesting_events[
                interesting_events[io._EventInputFileReader.COLUMN_CONTEXT] == context
            ]
            splits[context] = context_events[
                io._EventInputFileReader.COLUMN_TIME
            ].values
        return splits

    def _get_segment(
        self,
        trial: model.Trial,
        start_time: float,
        end_time: float,
        cycle_id: int,
        context: str,
    ) -> model.Trial:
        """Segments the trial data based on the start and end times.

        Args:
            trial: The trial to be segmented.
            start_time: The start time of the segment.
            end_time: The end time of the segment.
            cycle_id: The cycle id of the segment.
            context: The context of the segment.

        Returns:
            A new trial containing the segmented data.
        """

        trial_segment = model.Trial()
        # segment the data
        for category, data in trial.get_all_data().items():
            rate = int(data.attrs["rate"])
            dec_places = ga_math.get_decimal_places(1 / rate)

            segment = data.sel(time=slice(start_time, end_time))
            segment = segment.reset_index("time")

            times = (segment.time - start_time).to_numpy()
            times = np.round(times, dec_places)
            times = np.absolute(times)

            segment = segment.assign_coords(time=times)
            self._update_attrs(segment, start_time, end_time, cycle_id, context)
            trial_segment.add_data(category, segment)
        # segment the events
        trial_segment.events = self._segment_events(
            context, cycle_id, trial.events, start_time, end_time
        )
        return trial_segment

    @staticmethod
    def _segment_events(
        context: str,
        cycle_id: int,
        events: pd.DataFrame | None,
        start_time: float,
        end_time: float,
    ) -> pd.DataFrame:
        """Segments the events based on the start and end times.

        Args:
            context: The context of the segment.
            cycle_id: The cycle id of the segment.
            events: The events to be segmented.
            start_time: The start time of the segment.
            end_time: The end time of the segment.

        Returns:
            A DataFrame containing the segmented events.
        """
        if events is None:
            raise ValueError("Events are not set.")
        new_events = events[
            (events[io._EventInputFileReader.COLUMN_TIME] >= start_time)
            & (events[io._EventInputFileReader.COLUMN_TIME] <= end_time)
        ]
        new_events.loc[:, "time"] -= start_time
        new_events.attrs = {
            "end_time": end_time,
            "start_time": start_time,
            "context": context,
            "cycle_id": cycle_id,
            "used": 1,
        }
        return new_events

    @staticmethod
    def _update_attrs(
        segment: xr.DataArray, start_time, end_time, cycle_id: int, context: str
    ):
        """Updates the attributes of the segment based on the data.

        Updates time, and frames to relative values. Add additional information
        such as context, cycles_id and used. Based on the "used"-Flag cycles can be
        included or excluded in the analysis

        Args:
            segment: The segment to be updated.
            cycle_id: The cycle id of the segment.
            context: The context of the segment.
        """

        segment.attrs["start_time"] = start_time
        segment.attrs["end_time"] = end_time
        segment.attrs["cycle_id"] = cycle_id
        segment.attrs["context"] = context
        # netcdf can not handle booleans :(
        segment.attrs["used"] = 1



from abc import ABC, abstractmethod

import numpy as np
import pandas as pd
import xarray as xr
import logging

import gaitalytics.events as events
import gaitalytics.io as io
import gaitalytics.mapping as mapping
import gaitalytics.model as model
import gaitalytics.utils.linalg as linalg
import gaitalytics.utils.math as math
import gaitalytics.utils.mocap as mocap

logger = logging.getLogger(__name__)


class FeatureCalculation(ABC):
    """Base class for feature calculations.

    This class provides a common interface for calculating features.
    """

    def __init__(self, config: mapping.MappingConfigs, **kwargs):
        """Initializes a new instance of the BaseFeatureCalculation class.

        Args:
            config: The mapping configuration to use for the feature calculation.
            **kwargs: Currently not used.
        """
        self._config = config
        self._kwargs = kwargs

    @abstractmethod
    def calculate(self, trial: model.TrialCycles) -> xr.DataArray:
        """Calculate the features for a trial.

        Args:
            trial: The trial for which to calculate the features.

        Returns:
            An xarray DataArray containing the calculated features.
        """
        raise NotImplementedError


class CycleFeaturesCalculation(FeatureCalculation, ABC):
    def calculate(self, trial: model.TrialCycles) -> xr.DataArray:
        """Calculate the features for a trial.

        Calls the _calculate method for each cycle in the trial and combines
        results into a single DataArray.

        Args:
            trial: The trial for which to calculate the features.

        Returns:
            An xarray DataArray containing the calculated features.
        """
        results: list = []

        context_dim: list[str] = []
        for context, context_cycles in trial.get_all_cycles().items():
            context_results: list = []
            cycle_dim: list[int] = []
            for cycle_id, cycle in context_cycles.items():
                feature = self._calculate(cycle)
                context_results.append(feature)
                cycle_dim.append(cycle_id)

            context_dim.append(context)
            context_results = xr.concat(
                context_results, pd.Index(cycle_dim, name="cycle")
            )
            results.append(context_results)

        result = xr.concat(results, pd.Index(context_dim, name="context"))
        return result

    @abstractmethod
    def _calculate(self, trial: model.Trial) -> xr.DataArray:
        """Calculate the features for a trial.

        Args:
            trial: The trial for which to calculate the features.

        Returns:
            An xarray DataArray containing the calculated features.

        :meta public:
        """
        raise NotImplementedError

    @staticmethod
    def get_event_times(
        trial_events: pd.DataFrame | None,
    ) -> tuple[float, float, float, float, float]:
        """Checks the sequence of events in the trial and returns the times.

        Args:
            trial_events: The events to be checked and extracted.

        Returns:
            The times of the events. in following order
            [contra_fo, contra_fs, ipsi_fo, end_time]

        Raises:
            ValueError: If the sequence of events is incorrect.
        """
        if trial_events is None:
            raise ValueError("Trial does not have events.")
        curren_context = trial_events.attrs["context"]
        cycle_id = trial_events.attrs["cycle_id"]

        if len(trial_events) < 3:
            raise ValueError(
                f"Missing events in segment {curren_context} nr. {cycle_id}"
            )
        ipsi = trial_events[
            trial_events[io._EventInputFileReader.COLUMN_CONTEXT] == curren_context
        ]
        contra = trial_events[
            trial_events[io._EventInputFileReader.COLUMN_CONTEXT] != curren_context
        ]
        if len(ipsi) != 3:
            raise ValueError(f"Error events sequence {curren_context} nr. {cycle_id}")
        if len(contra) != 2:
            raise ValueError(f"Error events sequence {curren_context} nr. {cycle_id}")

        contra_fs = contra[
            contra[io._EventInputFileReader.COLUMN_LABEL] == events.FOOT_STRIKE
        ]
        contra_fo = contra[
            contra[io._EventInputFileReader.COLUMN_LABEL] == events.FOOT_OFF
        ]
        ipsi_fs = ipsi[
            ipsi[io._EventInputFileReader.COLUMN_LABEL] == events.FOOT_STRIKE
        ]
        ipsi_fo = ipsi[ipsi[io._EventInputFileReader.COLUMN_LABEL] == events.FOOT_OFF]

        ipsi_fs_time_start = ipsi_fs[io._EventInputFileReader.COLUMN_TIME].values[0]
        ipsi_fs_time_end = ipsi_fs[io._EventInputFileReader.COLUMN_TIME].values[1]
        ipsi_fo_time = ipsi_fo[io._EventInputFileReader.COLUMN_TIME].values[0]
        contra_fs_time = contra_fs[io._EventInputFileReader.COLUMN_TIME].values[0]
        contra_fo_time = contra_fo[io._EventInputFileReader.COLUMN_TIME].values[0]

        return (
            ipsi_fs_time_start,
            contra_fo_time,
            contra_fs_time,
            ipsi_fo_time,
            ipsi_fs_time_end,
        )

    @staticmethod
    def _create_result_from_dict(result_dict: dict) -> xr.DataArray:
        """Create a xarray DataArray from a dictionary.

        The dictionary should follow the format {feature: value}.
        In example:
        {
            "min": 1.0,
            "max": 2.0,
            "mean": 1.5,
            "median": 1.5,
            "std": 0.5,
        }

        Args:
            result_dict: The dictionary to create the DataArray from.

        Returns:
            An xarray DataArray containing the data from the dictionary.
        :meta public:
        """
        xr_dict = {
            "coords": {
                "feature": {"dims": "feature", "data": list(result_dict.keys())}
            },
            "data": list(result_dict.values()),
            "dims": "feature",
        }
        return xr.DataArray.from_dict(xr_dict)

    @staticmethod
    def _flatten_features(features: xr.DataArray) -> xr.DataArray:
        """Flatten the features into a single dimension.

        The dimension channel will be integrated into the features dim

        Args:
            features: The features to be flattened.

        Returns:
            The reshaped features.

        :meta public:
        """

        np_data = features.to_numpy()
        rs_data = np_data.reshape((np_data.shape[1] * np_data.shape[0]), order="C")

        feature = features.coords["feature"].values
        channel = features.coords["channel"].values
        new_labels = []
        for f in feature:
            for c in channel:
                new_labels.append(f"{c}_{f}")

        new_format = xr.DataArray(
            rs_data,
            coords={"feature": new_labels},
            dims=["feature"],
        )
        return new_format


class PointDependentFeature(CycleFeaturesCalculation, ABC):
    def _get_marker_data(
        self, trial: model.Trial, marker: mapping.MappedMarkers
    ) -> xr.DataArray:
        """Get the marker data for a trial.

        Args:
            trial: The trial for which to get the marker data.
            marker: The marker to get the data for.

        Returns:
            An xarray DataArray containing the marker data.

        :meta public:
        """
        return mocap.get_marker_data(trial, self._config, marker)

    def _get_sacrum_marker(self, trial: model.Trial) -> xr.DataArray:
        """Get the sacrum marker data for a trial.

        Try to get the sacrum marker data from the trial.
        If the sacrum marker not found calculate from posterior hip markers

        Args:
            trial: The trial for which to get the marker data.

        Returns:
            An xarray DataArray containing the sacrum marker data.

        :meta public:
        """
        return mocap.get_sacrum_marker(trial, self._config)

    def _get_progression_vector(self, trial: model.Trial) -> xr.DataArray:
        """Calculate the progression vector for a trial.

        The progression vector is the vector from the sacrum to the anterior hip marker.

        Args:
            trial: The trial for which to calculate the progression vector.

        Returns:
            An xarray DataArray containing the calculated progression vector.

        :meta public:
        """
        return mocap.get_progression_vector(trial, self._config)

    def _get_sagittal_vector(self, trial: model.Trial) -> xr.DataArray:
        """Calculate the sagittal vector for a trial.

        The sagittal vector is the vector normal to the sagittal plane. Note that this vector will always be pointing towards the right side of the body.

        Args:
            trial: The trial for which to calculate the sagittal vector.
        Returns:
            An xarray DataArray containing the calculated sagittal vector.
        :meta public:
        """
        progression_vector = self._get_progression_vector(trial)
        vertical_vector = xr.DataArray(
            [0, 0, 1], dims=["axis"], coords={"axis": ["x", "y", "z"]}
        )
        return linalg.get_normal_vector(progression_vector, vertical_vector)


class TimeSeriesFeatures(CycleFeaturesCalculation):
    """Calculate time series features for a trial.

    This class calculates following time series features for a trial.
    - min
    - max
    - mean
    - median
    - std
    """

    def _calculate(self, trial: model.Trial) -> xr.DataArray:
        """Calculate the time series features for a trial.

        Following features are calculated:
            - min
            - max
            - mean
            - median
            - std
            - amplitude

        Args:
            trial: The trial for which to calculate the features.

        Returns:
            An xarray DataArray containing the calculated features.
        """
        features = self._calculate_features(trial)
        return self._flatten_features(features)

    @staticmethod
    def _calculate_features(trial: model.Trial) -> xr.DataArray:
        """Calculate the time series features for a trial.

        Following features are calculated:
            - min
            - max
            - mean
            - median
            - std
            - amplitude


        Args:
            trial: The trial for which to calculate the features.

        Returns:
            An xarray DataArray containing the calculated features.
        """
        markers = trial.get_data(model.DataCategory.ANALYSIS)
        min_feat = markers.min(dim="time", skipna=True)
        max_feat = markers.max(dim="time", skipna=True)
        mean_feat = markers.mean(dim="time", skipna=True)
        median_feat = markers.median(dim="time", skipna=True)
        std_feat = markers.std(dim="time", skipna=True)
        amplitude_feat = max_feat - min_feat
        features = xr.concat(
            [min_feat, max_feat, mean_feat, median_feat, std_feat, amplitude_feat],
            pd.Index(
                ["min", "max", "mean", "median", "std", "amplitude"], name="feature"
            ),
        )

        return features


class PhaseTimeSeriesFeatures(TimeSeriesFeatures):
    """Calculate phase time series features for a trial.

    This class calculates following phase time series features for a trial.
        - stand_min
        - stand_max
        - stand_mean
        - stand_median
        - stand_std
        - stand_amplitude
        - swing_min
        - swing_max
        - swing_mean
        - swing_median
        - swing_std
        - swing_amplitude
    """

    def _calculate(self, trial: model.Trial) -> xr.DataArray:
        """Calculate the time series features for a trial by phase.

        Following features are calculated:
            - stand_min
            - stand_max
            - stand_mean
            - stand_median
            - stand_std
            - stand_amplitude
            - swing_min
            - swing_max
            - swing_mean
            - swing_median
            - swing_std
            - swing_amplitude

        Args:
            trial: The trial for which to calculate the features.

        Returns:
            An xarray DataArray containing the calculated features.
        """
        event_table = trial.events
        analysis_data = trial.get_data(model.DataCategory.ANALYSIS)

        context = analysis_data.attrs["context"]

        event_ipsi = event_table.loc[  # type: ignore
            event_table[io.C3dEventInputFileReader.COLUMN_CONTEXT] == context  # type: ignore
        ]
        event_ipsi_fo = event_ipsi.loc[
            event_table[io.C3dEventInputFileReader.COLUMN_LABEL] == events.FOOT_OFF  # type: ignore
        ]
        fo_time = round(event_ipsi_fo.time.values[0], 4)
        start_time = analysis_data.coords["time"].values[0]
        end_time = analysis_data.coords["time"].values[-1]

        stand_trial = self._create_phase_trial(trial, slice(start_time, fo_time))
        swing_trial = self._create_phase_trial(trial, slice(fo_time, end_time))
        stand_features = super()._calculate(stand_trial)
        stand_features = stand_features.assign_coords(
            feature=[f"{f}_stand" for f in stand_features.feature.values]
        )
        swing_features = super()._calculate(swing_trial)
        swing_features = swing_features.assign_coords(
            feature=[f"{f}_swing" for f in swing_features.feature.values]
        )

        return xr.concat([stand_features, swing_features], dim="feature")

    @staticmethod
    def _create_phase_trial(trial: model.Trial, time_slice: slice):
        """Create a trial containing only the data for a specific phase.

        Args:
            trial: The trial to create the phase trial from.
            time_slice: The time slice to extract the phase data from.
        """
        phase_trial = model.Trial()
        for data_category, data in trial.get_all_data().items():
            phase_trial.add_data(data_category, data.sel(time=time_slice))
        return phase_trial


class TemporalFeatures(CycleFeaturesCalculation):
    """Calculate temporal features for a trial.

    This class calculates following temporal features for a trial.
        - double_support
        - single_support
        - stance_duration_prec
        - swing_duration_prec
        - opposite_foot_off_prec
        - opposite_foot_contact_prec
        - stride_duration
        - stance_duration
        - cadence
    """

    def _calculate(self, trial: model.Trial) -> xr.DataArray:
        """Calculate the support times for a trial.

        Definitions of the temporal features
        Hollmann et al. 2011 (doi: 10.1016/j.gaitpost.2011.03.024)

        Args:
            trial: The trial for which to calculate the features.

        Returns:
            An xarray DataArray containing the calculated features.

        Raises:
            ValueError: If the sequence of events is incorrect.
        """
        trial_events = trial.events
        if trial_events is None:
            raise ValueError("Trial does not have events.")

        rel_event_times = self.get_event_times(trial_events)

        result_dict = self._calculate_supports(
            rel_event_times[1],
            rel_event_times[2],
            rel_event_times[3],
            rel_event_times[4],
        )
        result_dict["stance_duration_prec"] = rel_event_times[3] / rel_event_times[4]
        result_dict["opposite_foot_off_prec"] = rel_event_times[1] / rel_event_times[4]
        result_dict["opposite_foot_contact_prec"] = (
            rel_event_times[2] / rel_event_times[4]
        )
        result_dict["stride_duration"] = rel_event_times[4]
        result_dict["step_duration"] = rel_event_times[4] - rel_event_times[3]
        result_dict["swing_duration_prec"] = (
            (result_dict["step_duration"]) / rel_event_times[4]
        )
        result_dict["cadence"] = 60 / (rel_event_times[4] / 2)

        return self._create_result_from_dict(result_dict)

    @staticmethod
    def _calculate_supports(
        contra_fo_time: float,
        contra_fs_time: float,
        ipsi_fo_time: float,
        end_time: float,
    ) -> dict[str, float]:
        """Calculate the support times for a trial.

        Args:
            contra_fo_time: The time of the contra foot off event.
            contra_fs_time: The time of the contra foot strike event.
            ipsi_fo_time: The time of the ipsi foot off event.
            end_time: The end time of the trial.

        Returns:
            The calculated support times.
        """
        double_support = (contra_fo_time + (ipsi_fo_time - contra_fs_time)) / end_time
        single_support = (contra_fs_time - contra_fo_time) / end_time
        return {"double_support": double_support, "single_support": single_support}


class SpatialFeatures(PointDependentFeature):
    """Calculate spatial features for a trial.

    This class calculates following spatial features for a trial.
    - step_length
    - step_width
    - minimal_toe_clearance
    - AP_margin_of_stability
    - AP_base_of_support
    - AP_xcom
    - ML_margin_of_stability
    - ML_base_of_support
    - ML_xcom
    """

    def _calculate(self, trial: model.Trial) -> xr.DataArray:
        """Calculate the spatial features for a trial.

        Definitions of the spatial features:
        Step length & Step width: Hollmann et al. 2011 (doi: 10.1016/j.gaitpost.2011.03.024)
        Margin of stability: Jinfeng et al. 2021 (doi: 10.1152/jn.00091.2021), Curtze et al. 2024 (doi: 10.1016/j.jbiomech.2024.112045)
        Minimal toe clearance: Schulz 2017 (doi: 10.1016/j.jbiomech.2017.02.024)

        Args:
            trial: The trial for which to calculate the features.

        Returns:
            An xarray DataArray containing the calculated features.

        Raises:
            ValueError: If the trial does not have events.
        """

        if trial.events is None:
            raise ValueError("Trial does not have events.")

        marker_dict = self._select_markers_for_spatial_features(trial)

        results_dict = self._calculate_step_length(
            trial,
            marker_dict["ipsi_heel"],  # type: ignore
            marker_dict["contra_heel"],  # type: ignore
        )
        results_dict.update(
            self._calculate_step_width(
                trial,
                marker_dict["ipsi_heel"],  # type: ignore
                marker_dict["contra_heel"],  # type: ignore
            )
        )
        results_dict.update(
            self._calculate_stride_length(
                trial,
                marker_dict["ipsi_heel"],  # type: ignore
                marker_dict["contra_heel"],  # type: ignore
            )
        )

        toe_markers = [marker_dict["ipsi_toe_2"], marker_dict["ipsi_toe_5"]]

        results_dict.update(
            self._calculate_minimal_toe_clearance(
                trial,
                toe_markers,  # type: ignore
                marker_dict["ipsi_heel"],  # type: ignore
            )
        )

        try:
            results_dict.update(
                self._calculate_ap_margin_of_stability(
                    trial,
                    marker_dict["ipsi_heel"],  # type: ignore
                    marker_dict["contra_toe_2"],  # type: ignore
                    marker_dict["xcom"],
                )
            )

            results_dict.update(
                self._calculate_ml_margin_of_stability(
                    trial,
                    marker_dict["ipsi_ankle"],
                    marker_dict["contra_ankle"],
                    marker_dict["xcom"],
                )
            )
        except KeyError:
            logger.info(
                "Margin of stability markers are missing. No margin of stability calculated"
            )

        return self._create_result_from_dict(results_dict)

    @staticmethod
    def _select_markers_for_spatial_features(
        trial: model.Trial,
    ) -> dict[str, mapping.MappedMarkers]:
        """Select markers based on the trial's context (Right or Left). If some markers are missing, return them as None

        Args:
            trial: The trial object containing event attributes.

        Returns:
            A dictionary of markers based on context
        """
        if trial.events is not None and trial.events.attrs["context"] == "Right":
            ipsi_heel_marker = mapping.MappedMarkers.R_HEEL
            ipsi_toe_2_marker = mapping.MappedMarkers.R_TOE
            ipsi_toe_5_marker = mapping.MappedMarkers.R_TOE_2
            ipsi_ankle_marker = mapping.MappedMarkers.R_ANKLE

            contra_toe_2_marker = mapping.MappedMarkers.L_TOE
            contra_heel_marker = mapping.MappedMarkers.L_HEEL
            contra_ankle_marker = mapping.MappedMarkers.L_ANKLE

        else:
            ipsi_heel_marker = mapping.MappedMarkers.L_HEEL
            ipsi_toe_2_marker = mapping.MappedMarkers.L_TOE
            ipsi_toe_5_marker = mapping.MappedMarkers.L_TOE_2
            ipsi_ankle_marker = mapping.MappedMarkers.L_ANKLE

            contra_toe_2_marker = mapping.MappedMarkers.R_TOE
            contra_heel_marker = mapping.MappedMarkers.R_HEEL
            contra_ankle_marker = mapping.MappedMarkers.R_ANKLE

        xcom_marker = mapping.MappedMarkers.XCOM

        return {
            "ipsi_toe_2": ipsi_toe_2_marker,
            "ipsi_toe_5": ipsi_toe_5_marker,
            "ipsi_heel": ipsi_heel_marker,
            "ipsi_ankle": ipsi_ankle_marker,
            "contra_toe_2": contra_toe_2_marker,
            "contra_heel": contra_heel_marker,
            "contra_ankle": contra_ankle_marker,
            "xcom": xcom_marker,
        }

    def _calculate_step_length(
        self,
        trial: model.Trial,
        ipsi_marker: mapping.MappedMarkers,
        contra_marker: mapping.MappedMarkers,
    ) -> dict[str, np.ndarray]:
        """Calculate the step length for a trial.

        Args:
            trial: The trial for which to calculate the step length.
            ipsi_marker: The ipsi-lateral heel marker.
            contra_marker: The contra-lateral heel marker.

        Returns:
            The calculated step length.
        """

        event_times = self.get_event_times(trial.events)

        ipsi_heel = self._get_marker_data(trial, ipsi_marker).sel(
            time=event_times[-1], method="nearest"
        )
        contra_heel = self._get_marker_data(trial, contra_marker).sel(
            time=event_times[-1], method="nearest"
        )
        progress_axis = self._get_progression_vector(trial)
        progress_axis = linalg.normalize_vector(progress_axis)
        projected_ipsi = linalg.project_point_on_vector(ipsi_heel, progress_axis)
        projected_contra = linalg.project_point_on_vector(contra_heel, progress_axis)
        distance = linalg.calculate_distance(projected_ipsi, projected_contra).values
        return {"step_length": distance}

    def _calculate_step_width(
        self,
        trial: model.Trial,
        ipsi_marker: mapping.MappedMarkers,
        contra_marker: mapping.MappedMarkers,
    ) -> dict[str, np.ndarray]:
        """Calculate the step width for a trial.

        Args:
            trial: The trial for which to calculate the step width.
            ipsi_marker: The ipsi-lateral heel marker.
            contra_marker: The contra-lateral heel marker.

        Returns:
            The calculated step width in a dict.
        """

        event_times = self.get_event_times(trial.events)
        contra_heel = self._get_marker_data(trial, contra_marker)
        contra_vector = contra_heel.sel(
            time=event_times[2], method="nearest"
        ) - contra_heel.sel(time=event_times[0], method="nearest")

        ipsi_heel = self._get_marker_data(trial, ipsi_marker).sel(
            time=event_times[-1], method="nearest"
        )

        norm_vector = linalg.normalize_vector(contra_vector)
        projected_ipsi = linalg.project_point_on_vector(ipsi_heel, norm_vector)
        distance = linalg.calculate_distance(ipsi_heel, projected_ipsi).values

        return {"step_width": distance}

    def _calculate_stride_length(
        self,
        trial: model.Trial,
        ipsi_marker: mapping.MappedMarkers,
        contra_marker: mapping.MappedMarkers,
    ) -> dict[str, np.ndarray]:
        """Calculate the stride length for a trial. It is computed as the two consecutive step lengths constituting the gait cycle.

        Args:
            trial: The trial for which to calculate the stride length.
            ipsi_marker: The ipsi-lateral heel marker.
            contra_marker: The contra-lateral heel marker.

        Returns:
            The calculated stride length.
        """

        event_times = self.get_event_times(trial.events)
        progress_axis = self._get_progression_vector(trial)
        progress_axis = linalg.normalize_vector(progress_axis)

        total_distance = None

        # Add the distance of the ipsi and contra heel step length
        for event_time in [event_times[2], event_times[-1]]:
            ipsi_heel = self._get_marker_data(trial, ipsi_marker).sel(
                time=event_time, method="nearest"
            )
            contra_heel = self._get_marker_data(trial, contra_marker).sel(
                time=event_time, method="nearest"
            )

            projected_ipsi = linalg.project_point_on_vector(ipsi_heel, progress_axis)
            projected_contra = linalg.project_point_on_vector(
                contra_heel, progress_axis
            )

            distance = linalg.calculate_distance(
                projected_ipsi, projected_contra
            ).values
            if total_distance is None:
                total_distance = distance
            else:
                total_distance += distance

        if total_distance is None:
            total_distance = np.empty(1)

        return {"stride_length": total_distance}

    def _calculate_minimal_toe_clearance(
        self,
        trial: model.Trial,
        ipsi_toe_markers: list[mapping.MappedMarkers],
        ipsi_heel_marker: mapping.MappedMarkers,
    ) -> dict[str, np.ndarray]:
        """Calculate the minimal toe clearance for a trial. <br />
        Toe clearance is computed for all toe markers passed, only the minimal is returned. <br />
        Addition toe markers can be passed to improve the accuracy of the calculation. <br />
        If the minimal toe clearance cannot be calculated, an empty array is returned. <br />

        Args:
            trial: The trial to compute the minimal toe clearance for
            ipsi_toe_markers: The ipsi-lateral toe markers
            ipsi_heel_marker (mapping.MappedMarkers): The ipsi-lateral heel marker

        Returns:
            dict[str, np.ndarray]: The calculated minimal toe clearance in a dict

        Raises:
            ValueError: If no toe markers are found for minimal toe clearance calculation
        """
        event_times = self.get_event_times(trial.events)

        ipsi_heel = self._get_marker_data(trial, ipsi_heel_marker).sel(
            time=slice(event_times[3], event_times[4])
        )
        ipsi_toes = []
        ipsi_toe_velocities = None

        # Get all ipsi toe marker positions and calculate the mean velocity of them
        for ipsi_toe_marker in ipsi_toe_markers:
            try:
                ipsi_toe = self._get_marker_data(trial, ipsi_toe_marker).sel(
                    time=slice(event_times[3], event_times[4])
                )
                ipsi_toes.append(ipsi_toe)
                if ipsi_toe_velocities is None:
                    ipsi_toe_velocities = linalg.calculate_speed_norm(ipsi_toe)
                else:
                    ipsi_toe_velocities += linalg.calculate_speed_norm(ipsi_toe)
            except KeyError:
                pass

        if ipsi_toe_velocities is None:
            raise ValueError(
                "No toe markers found for minimal toe clearance calculation"
            )

        ipsi_toe_velocities /= len(ipsi_toes)

        minimal_toe_clearance = np.inf
        for ipsi_toe in ipsi_toes:
            i = self._find_mtc_index(ipsi_toe, ipsi_heel, ipsi_toe_velocities)
            if i is not None:
                clearance = ipsi_toe.sel(axis="z")[i].values
                if clearance < minimal_toe_clearance:
                    minimal_toe_clearance = clearance  # type: ignore

        return {
            "minimal_toe_clearance": minimal_toe_clearance  # type: ignore
            if minimal_toe_clearance != np.inf
            else np.nan
        }

    @staticmethod
    def _find_mtc_index(
        toe_position: xr.DataArray, heel_position: xr.DataArray, toes_vel: xr.DataArray
    ):
        """Find the time corresponding to minimal toe clearance of a specific toe.
        Valid minimal toe clearance point must pass conditions
        defined in Schulz 2017 (doi: 10.1016/j.jbiomech.2017.02.024)
        Args:
            toe_position: A DataArray containing positions of the toe
            heel_position: A DataArray containing positions of the heel
            toes_vel: A DataArray containing the mean toes velocity at each point in time
        Returns:
            The time corresponding to minimal toe clearance for the input toe.
        """
        toes_vel_up_quant = np.quantile(toes_vel, 0.5)
        toe_z = toe_position.sel(axis="z")
        heel_z = heel_position.sel(axis="z")

        # Check conditions according to Schulz 2017
        mtc_i = math.find_local_minimas(toe_z)
        mtc_i = [i for i in mtc_i if toes_vel[i] >= toes_vel_up_quant]
        mtc_i = [i for i in mtc_i if toe_z[i] <= heel_z[i]]

        return None if not mtc_i else min(mtc_i, key=lambda i: toe_z[i])  # type: ignore

    def _calculate_ap_margin_of_stability(
        self,
        trial: model.Trial,
        ipsi_heel_marker: mapping.MappedMarkers,
        contra_toe_marker: mapping.MappedMarkers,
        xcom_marker: mapping.MappedMarkers,
    ) -> dict[str, np.ndarray]:
        """Calculate the anterio-posterior margin of stability at heel strike. Result should be interpreted according to Curtze et al. (2024)
        Args:
            trial: The trial for which to calculate the AP margin of stability
            ipsi_heel_marker: The ipsi-lateral heel marker
            contra_marker: The contra-lateral toe marker
            xcom_marker: The extrapolated center of mass marker

        Returns:
            dict: A dictionary containing:
                - "AP_margin_of_stability": The calculated anterio-posterior margin of stability.
                - "AP_base_of_support": The calculated anterio-posterior base of support.
                - "AP_XCOM": The calculated anterio-posterior position of the extrapolated center of mass relative to the back foot.
        """
        event_times = self.get_event_times(trial.events)

        ipsi_heel = self._get_marker_data(trial, ipsi_heel_marker).sel(
            time=event_times[0], method="nearest"
        )
        contra_toe = self._get_marker_data(trial, contra_toe_marker).sel(
            time=event_times[0], method="nearest"
        )
        xcom = self._get_marker_data(trial, xcom_marker).sel(
            time=event_times[0], method="nearest"
        )

        progress_axis = self._get_progression_vector(trial)
        progress_axis = linalg.normalize_vector(progress_axis)

        front_marker = linalg.get_point_in_front(ipsi_heel, contra_toe, progress_axis)
        back_marker = linalg.get_point_behind(ipsi_heel, contra_toe, progress_axis)

        bos_vect = front_marker - back_marker
        xcom_vect = xcom - back_marker

        bos_proj = abs(linalg.signed_projection_norm(bos_vect, progress_axis))
        xcom_proj = linalg.signed_projection_norm(xcom_vect, progress_axis)
        mos = bos_proj - xcom_proj

        return {
            "AP_margin_of_stability": mos.to_numpy(),
            "AP_base_of_support": bos_proj.to_numpy(),
            "AP_xcom": xcom_proj.to_numpy(),
        }

    def _calculate_ml_margin_of_stability(
        self,
        trial: model.Trial,
        ipsi_ankle_marker: mapping.MappedMarkers,
        contra_ankle_marker: mapping.MappedMarkers,
        xcom_marker: mapping.MappedMarkers,
    ) -> dict[str, np.ndarray]:
        """Calculate the medio-lateral margin of stability at heel strike. Result should be interpreted according to Curtze et al. (2024)
        Args:
            trial: The trial for which to calculate the ml margin of stability
            ipsi_toe_marker: The ipsi-lateral lateral ankle marker
            contra_marker: The contra-lateral lateral ankle marker
            xcom_marker: The extrapolated center of mass marker

        Returns:
            dict: A dictionary containing:
                - "ML_margin_of_stability": The calculated medio-lateral margin of stability.
                - "ML_base_of_support": The calculated medio-lateral base of support.
                - "ML_xcom": The calculated medio-lateral position of the extrapolated center of mass relative to the back foot.
        """
        event_times = self.get_event_times(trial.events)

        ipsi_ankle = self._get_marker_data(trial, ipsi_ankle_marker).sel(
            time=event_times[0], method="nearest"
        )
        contra_ankle = self._get_marker_data(trial, contra_ankle_marker).sel(
            time=event_times[0], method="nearest"
        )
        xcom = self._get_marker_data(trial, xcom_marker).sel(
            time=event_times[0], method="nearest"
        )

        sagittal_axis = self._get_sagittal_vector(trial)
        sagittal_axis = linalg.normalize_vector(sagittal_axis)

        if trial.events is not None and trial.events.attrs["context"] == "Left":
            # Rotate sagittal axis so it points towards the left side of the body
            sagittal_axis = -sagittal_axis

        # Lateral is the furthest point in the direction of the sagittal axis
        lateral_point = linalg.get_point_in_front(
            ipsi_ankle, contra_ankle, sagittal_axis
        )
        # Medial is the closest point in the direction of the sagittal axis
        medial_point = linalg.get_point_behind(ipsi_ankle, contra_ankle, sagittal_axis)

        bos_vect = lateral_point - medial_point
        xcom_vect = xcom - medial_point

        bos_proj = abs(linalg.signed_projection_norm(bos_vect, sagittal_axis))
        xcom_proj = linalg.signed_projection_norm(xcom_vect, sagittal_axis)
        mos = bos_proj - xcom_proj

        return {
            "ML_margin_of_stability": mos.to_numpy(),
            "ML_base_of_support": bos_proj.to_numpy(),
            "ML_xcom": xcom_proj.to_numpy(),
        }



from functools import wraps
from inspect import signature, Parameter
from pathlib import Path

import pandas as pd
import xarray as xr

import gaitalytics.events as events
import gaitalytics.features as features
import gaitalytics.io as io
import gaitalytics.mapping as mapping
import gaitalytics.model as model
import gaitalytics.normalisation as normalisation
import gaitalytics.segmentation as segmentation


class _PathConverter:
    """A decorator to convert Path | str annotations to Path objects.

    This decorator is used to convert parameters with Path | str annotations
    to Path objects.
    """

    def __init__(self, func):
        wraps(func)(self)
        self.func = func
        self.sig = signature(func)

    def __call__(self, *args, **kwargs):
        bound_arguments = self.sig.bind(*args, **kwargs)
        bound_arguments.apply_defaults()

        new_args = []
        for name, value in bound_arguments.arguments.items():
            param: Parameter = self.sig.parameters[name]
            # Check if the annotation is Path | str
            if param.annotation == Path | str or param.annotation == Path | str | None:
                if isinstance(value, (Path, str)):
                    value = Path(value)
            if param.kind in [
                param.VAR_POSITIONAL,
                param.POSITIONAL_ONLY,
                param.POSITIONAL_OR_KEYWORD,
            ]:
                new_args.append(value)
            else:
                bound_arguments.arguments[name] = value

        return self.func(*new_args, **bound_arguments.kwargs)


@_PathConverter
def load_config(config_path: Path | str) -> mapping.MappingConfigs:
    """Loads the mapping configuration file.

    Args:
        config_path: The path to the configuration file.

    Returns:
        A MappingConfigs object.
    """
    return mapping.MappingConfigs(config_path)  # type: ignore


@_PathConverter
def load_c3d_trial(
    c3d_file: Path | str, configs: mapping.MappingConfigs
) -> model.Trial:
    """Loads a Trial from a c3d file.

    Be aware that all the required data for the trial must be present in the c3d file.
    i.e. markers, analogs, events, etc.

    Args:
        c3d_file: The path to the c3d file.
        configs: The mapping configurations

    Returns:
        A Trial object.
    """
    markers = io.MarkersInputFileReader(c3d_file).get_markers()  # type: ignore
    analogs = io.AnalogsInputFileReader(c3d_file).get_analogs()  # type: ignore
    analysis = io.AnalysisInputReader(c3d_file, configs).get_analysis()  # type: ignore
    event_table = io.C3dEventInputFileReader(c3d_file).get_events()  # type: ignore

    trial = model.Trial()
    trial.add_data(model.DataCategory.MARKERS, markers)
    trial.add_data(model.DataCategory.ANALOGS, analogs)
    trial.add_data(model.DataCategory.ANALYSIS, analysis)
    trial.events = event_table

    return trial


def detect_events(
    trial: model.Trial,
    config: mapping.MappingConfigs,
    method: type[events.BaseEventDetection] = events.MarkerEventDetection,
    **kwargs,
) -> pd.DataFrame:
    """Detects the events in the trial.

    Args:
        trial: The trial to detect the events for.
        config: The mapping configurations
        method: The class to use for detecting the events.
                Currently, only "Marker" is supported, which implements
                the method from Zenis et al. 2008.
                Default is "Marker".
        **kwargs:
            - height: The height of peaks. Default = None
            - threshold: The threshold of peaks. Default = None
            - distance: The min distance in frames between events. Default = None
            - rel_height: The relative height of peak. Default = 0.5

    Returns:
        A DataFrame containing the detected events.

    """

    method_obj = method(config, **kwargs)

    event_table = method_obj.detect_events(trial)
    return event_table


def check_events(event_table: pd.DataFrame, method: str = "sequence"):
    """Checks the events in the trial.

    Args:
        event_table: The event table to check.
        method: The method to use for checking the events.
                Currently, only supports "sequence" which checks the sequence of events
                in terms of context and label. Default is "sequence".

    Returns:
        The trial with the checked events.

    Raises:
        ValueError: If the event sequence is not correct or the method is not supported.
    """
    match method:
        case "sequence":
            checker = events.SequenceEventChecker()
        case _:
            raise ValueError(f"Unsupported method: {method}")
    good, errors = checker.check_events(event_table)
    if not good:
        raise ValueError(f"Event sequence is not correct: {errors}")


@_PathConverter
def write_events_to_c3d(
    c3d_path: Path | str,
    event_table: pd.DataFrame,
    output_path: Path | str | None = None,
):
    """Writes the events to the c3d file.

    Args:
        c3d_path: The path to the original c3d file.
        event_table: The DataFrame containing the events.
        output_path: The path to write the c3d file with the events.
                     If None, the original file will be overwritten.
    """
    io.C3dEventFileWriter(c3d_path).write_events(event_table, output_path)  # type: ignore


def segment_trial(trial: model.Trial, method: str = "HS") -> model.TrialCycles:
    """Segments the trial into cycles

    Args:
        trial: The trial to segment.
        method: The method to use for segmenting the trial.
                Currently, only supports "HS" which segments the trial based on heel strikes.
                Default is "HS".

    Returns:
        The trial with the segmented data.
    """
    match method:
        case "HS":
            method_obj = segmentation.GaitEventsSegmentation()
        case "TO":
            method_obj = segmentation.GaitEventsSegmentation(events.FOOT_OFF)
        case _:
            raise ValueError(f"Unsupported method: {method}")

    trial_cycles = method_obj.segment(trial)
    return trial_cycles


def time_normalise_trial(
    trial: model.Trial | model.TrialCycles, method: str = "linear", **kwargs
) -> model.Trial | model.TrialCycles:
    """Normalises the time in the trial.

    Args:
        trial: The trial to normalise the time for.
        method: The method to use for normalising the time. Currently, only supports
                "linear" which normalises the time linearly. Default is "linear".
        **kwargs:
            - n_frames: The number of frames to normalise the data to.

    Returns:
        The trial with the normalised time.
    """
    match method:
        case "linear":
            normaliser = normalisation.LinearTimeNormaliser(**kwargs)
        case _:
            raise ValueError(f"Unsupported method: {method}")

    trial = normaliser.normalise(trial)
    return trial


def calculate_features(
    trial: model.TrialCycles,
    config: mapping.MappingConfigs,
    methods: list | tuple = (
        features.TimeSeriesFeatures,
        features.PhaseTimeSeriesFeatures,
        features.TemporalFeatures,
        features.SpatialFeatures,
    ),
    **kwargs,
) -> xr.DataArray:
    """Calculates the features of the trial.

    Args:
        trial: The trial to calculate the features for.
        config: The mapping configurations
        methods: Class objects of the feature calculation methods to use.
        **kwargs: Currently not used.

    Returns:
        The trial with the calculated features.
    """
    method_objs = _create_feature_methods(methods, config, **kwargs)
    feature_list = []
    for method in method_objs:
        feature_list.append(method.calculate(trial))
    feature_array = xr.concat(feature_list, dim="feature")
    return feature_array


def _create_feature_methods(
    methods: list[type] | tuple[type], config: mapping.MappingConfigs, **kwargs
) -> list[features.FeatureCalculation]:
    """Checks the feature calculation methods.

    Args:
        methods: The list of feature calculation methods to use.
        config: The mapping configurations
        **kwargs: Currently not used.

    Returns:
        A list of the feature calculation method objects.
    """
    method_objects = []
    for method in methods:
        if not issubclass(method, features.FeatureCalculation):
            raise ValueError(f"Unsupported method: {method}")
        method_objects.append(method(config, **kwargs))
    return method_objects


@_PathConverter
def export_trial(
    trial: model.Trial | model.TrialCycles,
    output_path: Path | str,
    method: str = "netcdf",
):
    """Exports the trial to a c3d file.

    This function will create a folder and save the trial as NetCDF files.
    Depending on the trial following files will be written:

    - markers.nc
    - analogs.nc
    - analysis.nc
    - events.nc

    Args:
        trial: The trial to export.
        output_path: The path to write the c3d file.
        method: The method to use for exporting the trial.
                Currently, only supports "netcdf".
    """
    match method:
        case "netcdf":
            io.NetCDFTrialExporter(output_path).export_trial(trial)  # type: ignore
        case _:
            raise ValueError(f"Unsupported method: {method}")



from gaitalytics import api

config = api.load_config("./tests/pig_config.yaml")
trial = api.load_c3d_trial("./tests/test_small.c3d", config)
trial_segmented = api.segment_trial(trial)
features = api.calculate_features(trial_segmented, config)



import decimal


def get_decimal_places(number: float) -> int:
    """Get the number of decimal places in a number.

    Args:
        number: The number to get the decimal places for.

    Returns:
        The number of decimal places in the number.
    """
    places = decimal.Decimal(str(number)).as_tuple().exponent
    if type(places) is not int:
        raise ValueError("The number of decimal places must be an integer.")

    return abs(places)


def find_local_minimas(arr):
    """
    Finds the indices of local minima in an array. A point is considered a local minima
    if it is smaller than the two points before and after it.

    Args:
        arr: The array to find minimas in

    Returns:
        A list of indices where the local minima are located.
    """
    local_minimas = []
    for i in range(2, len(arr) - 2):
        if (
            arr[i] < arr[i - 2]
            and arr[i] < arr[i - 1]
            and arr[i] < arr[i + 1]
            and arr[i] < arr[i + 2]
        ):
            local_minimas.append(i)

    return local_minimas



import xarray as xr
import numpy as np


def calculate_distance(
    point_a: xr.DataArray,
    point_b: xr.DataArray,
) -> xr.DataArray:
    """Calculate the distance between two points.

    Args:
        point_a: The first point.
        point_b: The second point.

    Returns:
        An xarray DataArray containing the calculated distance.
    """
    return (point_b - point_a).meca.norm(dim="axis")


def project_point_on_vector(point: xr.DataArray, vector: xr.DataArray) -> xr.DataArray:
    """Project a point onto a vector.

    Args:
        point: The point to project.
        vector: The vector to project onto.

    Returns:
        An xarray DataArray containing the projected point.
    """
    return vector * point.dot(vector, dim="axis")


def signed_projection_norm(vector: xr.DataArray, onto: xr.DataArray) -> xr.DataArray:
    """Compute the signed norm of the projection of a vector onto another vector. <br>
    If the projection is in the same direction as the onto vector, the norm is positive. <br>
    If the projection is in the opposite direction, the norm is negative. <br>

    Args:
        vector: The vector to be projected.
        onto: The vector to project onto.

    Returns:
        An xarray DataArray containing the signed norm of the projected vector.
    """
    projection = onto * vector.dot(onto, dim="axis") / onto.dot(onto, dim="axis")
    projection_norm = projection.meca.norm(dim="axis")
    sign = xr.where(vector.dot(onto, dim="axis") > 0, 1, -1)
    sign = xr.where(vector.dot(onto, dim="axis") == 0, 0, sign)
    return projection_norm * sign


def get_normal_vector(vector1: xr.DataArray, vector2: xr.DataArray) -> xr.DataArray:
    """Create a vector with norm = 1 normal to two other vectors.

    Args:
        vector1: The first vector to be normal to.
        vector2: The second vector to be normal to.

    Returns:
        An xarray DataArray containing the normal vector.
    """
    normal_vector = xr.DataArray(
        np.cross(vector1.values, vector2.values),
        dims=vector1.dims,
        coords=vector1.coords,
    )
    return normalize_vector(normal_vector)


def normalize_vector(vector: xr.DataArray) -> xr.DataArray:
    """Normalize a vector.

    Args:
        vector: The vector to normalize.

    Returns:
        An xarray DataArray containing the normalized vector.
    """
    return vector / vector.meca.norm(dim="axis")


def calculate_speed_norm(position: xr.DataArray, dt: float = 0.01) -> xr.DataArray:
    """
    Compute the speed from a 3xN position data array obtained with constant sampling rate

    Args:
        position: A 3xN xarray.DataArray where each row corresponds to an axis (x, y, z),
                          and each column represents a time point (positions in space).
        dt: Time interval between samples.

    Returns:
        An np array with the speed values.
    """
    velocity_squared_sum = sum(
        (np.diff(position.sel(axis=axis).values) / dt) ** 2
        for axis in position.coords["axis"]
    )
    speed_values = np.sqrt(velocity_squared_sum)
    speed_values = np.append(speed_values, speed_values[-1])

    return xr.DataArray(
        speed_values, dims=["time"], coords={"time": position.coords["time"]}
    )


def get_point_in_front(
    point_a: xr.DataArray, point_b: xr.DataArray, direction_vector: xr.DataArray
) -> xr.DataArray:
    """Determine which point is in front of the other according to the direction vector.

    Args:
        point_a: The first point.
        point_b: The second point.
        direction_vector: The direction vector.

    Returns:
        The point that is in front according to the direction vector.
    """
    direction_vector = direction_vector / direction_vector.meca.norm(dim="axis")
    vector_b_to_a = point_a - point_b
    signed_distance = vector_b_to_a.dot(direction_vector, dim="axis")

    return point_a if signed_distance > 0 else point_b


def get_point_behind(
    point_a: xr.DataArray, point_b: xr.DataArray, direction_vector: xr.DataArray
) -> xr.DataArray:
    """Determine which point is behind the other according to the direction vector.

    Args:
        point_a: The first point.
        point_b: The second point.
        direction_vector: The direction vector.

    Returns:
        The point that is behind according to the direction vector.
    """
    direction_vector = direction_vector / direction_vector.meca.norm(dim="axis")
    vector_b_to_a = point_a - point_b
    signed_distance = vector_b_to_a.dot(direction_vector, dim="axis")

    return point_b if signed_distance > 0 else point_a



from gaitalytics import api

config = api.load_config("./tests/pig_config.yaml")
trial = api.load_c3d_trial("./tests/test_small.c3d", config)
events = api.detect_events(trial, config)

try:
    api.check_events(trial.events)
except ValueError as e:
    print(e)

api.write_events_to_c3d("./tests/test_small.c3d", events,
                        "./out/test_small_events.c3d")




"""This module provides classes for structuring, storing and loading trial data."""

from abc import ABC, abstractmethod
from enum import Enum
from pathlib import Path

import h5netcdf as netcdf
import pandas as pd
import xarray as xr


class DataCategory(Enum):
    """Enum class for the array categories.

    This class provides the categories for the data arrays.


    Attributes:
        MARKERS: The marker's category.
        ANALOGS: The analog's category.
        ANALYSIS: The analysis category.
    """

    MARKERS: str = "markers"
    ANALOGS: str = "analogs"
    ANALYSIS: str = "analysis"


class BaseTrial(ABC):
    """Abstract base class for trials.

    This class provides a common interface for trials to load and save data.
    """

    @abstractmethod
    def _to_hdf5(self, file_path: Path, base_group: str | None = None):
        """Local implementation of the to_hdf5 method.

        Args:
            file_path: The path to the HDF5 file.
            base_group: The base group to save the data.
            If None, the data will be saved in the root of the file.
            Default = None
        """
        raise NotImplementedError

    def to_hdf5(self, file_path: Path, base_group: str | None = None):
        """Saves the trial data to an HDF5 file.

        Args:
            file_path: The path to the HDF5 file.
            base_group: The base group to save the data.
            If None, the data will be saved in the root of the file.
            Default = None

        Raises:
            FileExistsError: If the file already exists.
            ValueError: If the trial is a segmented trial and
            the file path is a single file.
            ValueError: If the trial is a trial and the file path is a folder.
        """
        if file_path.exists():
            raise FileExistsError(f"{file_path} already exists.")
        elif type(self) is TrialCycles and file_path.suffix:
            raise ValueError("Cannot save a segmented trial in a single file.")
        elif type(self) is Trial and not file_path.suffix:
            raise ValueError("Cannot save a trial in folder")

        paths, data, groups = self._to_hdf5(file_path, base_group)
        if len(data) > 0:
            if file_path.suffix:
                file_path.parent.mkdir(parents=True, exist_ok=True)
            else:
                file_path.mkdir(parents=True, exist_ok=True)
            xr.save_mfdataset(data, paths, groups=groups, mode="a", engine="h5netcdf")
        else:
            raise ValueError("No data to save.")


class Trial(BaseTrial):
    """Represents a trial.

    A trial is a collection of data arrays (typically markers & analogs) and events.
    """

    def __init__(self):
        """Initializes a new instance of the Trial class."""
        self._data: dict[DataCategory, xr.DataArray] = {}
        self._events: pd.DataFrame | None = None

    @property
    def events(self) -> pd.DataFrame | None:
        """Gets the events in the trial.

        Returns:
            pd.DataFrame: A pandas DataFrame containing the events if present.
            None: If no events are present.
        """
        return self._events

    @events.setter
    def events(self, events: pd.DataFrame):
        """Sets the events in the trial.

        Args:
            events: The events to be set.
        """
        self._events = events

    def add_channel(self, category: DataCategory, data: xr.DataArray):
        """Adds a channel to the trial.

        Args:
            category: The category of the data.
            data: The data array to be added.
        """
        if category in self._data:
            self._data[category] = xr.concat(
                [self._data[category], data], dim="channel"
            )
        else:
            self._data[category] = data

    def add_data(self, category: DataCategory, data: xr.DataArray):
        """Adds data to the trial.

        Args:
            category: The category of the data.
            data: The data array to be added.
        """
        if category in self._data:
            self._data[category] = xr.concat([self._data[category], data], dim="time")
        else:
            self._data[category] = data

    def get_data(self, category: DataCategory) -> xr.DataArray:
        """Gets the data from the trial.

        Args:
            category: The category of the data.

        Returns:
            The data array.
        """
        return self._data[category]

    def get_all_data(self) -> dict[DataCategory, xr.DataArray]:
        """Gets all data from the trial.

        Returns:
            A dictionary containing the data arrays.
        """
        return self._data

    def _to_hdf5(self, file_path: Path, base_group: str | None = None):
        """Saves trial into an HDF5 file.

        Structure:

        - root
            - markers
                - xarray.DataArray
            - analogs
                - xarray.DataArray
            - events
                - xarray.Dataset

        Args:
            file_path: The path to the HDF5 file.
            base_group: The base group to save the data.
                If None, the data will be saved in the root of the file.
                Default = ""
        """
        if base_group is None:
            base_group = ""
        else:
            base_group = f"{base_group}"

        groups = []
        data = []
        paths = []
        # Gather all data
        if self.get_all_data() is not None and len(self.get_all_data()) > 0:
            groups += [
                f"{base_group}{category.value}"
                for category in self.get_all_data().keys()
            ]
            data += [data.to_dataset() for data in self.get_all_data().values()]
            paths += [file_path for _ in groups]

        if self.events is not None:
            groups.append(f"{base_group}events")
            data.append(self.events.to_xarray())
            paths.append(file_path)

        return paths, data, groups


class TrialCycles(BaseTrial):
    """Represents a segmented trial."""

    def __init__(self):
        """Initializes a new instance of the SegmentedTrial class."""
        self._cycles: dict[str, dict[int, Trial]] = {}

    def add_cycle(self, context: str, cycle_id: int, segment: Trial):
        """Adds a Cycle to the segmented trial.

        Args:
            context: The context of the cycle.
            cycle_id: The id of the cycle.
            segment: The segment to be added.
        """
        if context not in self._cycles.keys():
            self._cycles[context] = {}

        self._cycles[context][cycle_id] = segment

    def get_cycle(self, context: str, cycle_id: int) -> Trial:
        """Gets a cycle from the segmented trial.

        Args:
            context: The context of the cycle.
            cycle_id: The id of the cycle.

        Raises:
            KeyError: If the cycle does not exist.
        """
        return self._cycles[context][cycle_id]

    def get_all_cycles(self) -> dict[str, dict[int, Trial]]:
        """Gets all cycles from the segmented trial.

        Returns:
            A nexted dictionary containing the cycles.
            Whereas the key of the first dictionary is the context and the second
            key is the cycle number.
        """
        return self._cycles

    def get_cycles_per_context(self, context: str) -> dict[int, Trial]:
        """Gets all cycles from the segmented trial for a specific context.

        Args:
            context: The context of the cycles.

        Returns:
            A dictionary containing the cycles
            for the specified context.
        """
        return self._cycles[context]

    def _to_hdf5(self, file_path: Path, base_group: str | None = None):
        """Recursively saves the segmented trial data to an HDF5 file.

        Unfortunately, writing a huge a mount of separate arrays in a single file
        is not efficient at the moment.
        So the data is saved in separate files.

        Structure example of GaitEventsSegmentation:
        /folder
            - 0.h5 (cycle_id)
                - Left
                    - markers
                        - xarray.DataArray
                    - analogs
                        - xarray.DataArray
                    - events
                        - xarray.DataSet
                - Right
                    - markers
                        - xarray.DataArray
                    - analogs
                        - xarray.DataArray
                    - events
                        - xarray.DataSet
            - ...

        Args:
            file_path: The path to the HDF5 file.
            base_group: The base group to save the data.
                If None, the data will be saved in the root of the file.
        """
        if base_group is None:
            base_group = "/"

        groups = []
        data = []
        paths = []

        for context, cycles in self.get_all_cycles().items():
            for cycle_id, cycle in cycles.items():
                new_file = (file_path / f"{cycle_id}").with_suffix(".h5")
                new_base_group = f"{base_group}{context}/"
                seg_path, seg_data, seg_groups = cycle._to_hdf5(
                    new_file, base_group=new_base_group
                )
                groups += seg_groups
                data += seg_data
                paths += seg_path

        return paths, data, groups


def trial_from_hdf5(file_path: Path) -> Trial | TrialCycles:
    """Loads trial data from an HDF5 file.

    Can be used to load a single trial or a segmented trial after saving it.

    Args:
        file_path: The path to the HDF5 file or folder with the expected structure.

    Returns:
        Trial: A new instance of the Trial class if file_path is a single file.
        TrialCycles: A new instance of the TrialCycles class if file_path is a folder.
    """
    trial: Trial | TrialCycles
    if not file_path.exists():
        raise FileNotFoundError(f"File {file_path} does not exist.")
    elif file_path.suffix:
        trial = _load_trial_file(file_path)
    else:
        trial = _load_segmented_trial_file(file_path)

    return trial


def _load_segmented_trial_file(file_path: Path) -> TrialCycles:
    """Loads a segmented trial from a folder containing HDF5 files.

    Following structure is expected:
    - folder
        - 0.h5 (cycle_id)
            - Left (context)
                - markers
                    - xarray.DataArray
                - analogs
                    - xarray.DataArray
                - events
                    - xarray.Dataset
            - Right (context)
                ...
        - ...

    Args:
        file_path: folder path containing the HDF5 files.

    Returns:
        A new instance of the TrialCycles class.

    """
    trial_cycles = TrialCycles()

    for file in file_path.glob("**/*.h5"):
        with netcdf.File(str(file), "r") as f:
            cycle_id = file.name.replace(".h5", "")
            for context in f.groups.keys():
                trial = _load_trial(f[context], file)
                trial_cycles.add_cycle(context, int(cycle_id), trial)
    return trial_cycles


def _load_trial_file(file_path: Path) -> Trial:
    """Loads a trial from an HDF5 file.

    Following structure is expected:
    - file_path (hdf5 file)
        - Left (context)
            - markers
                - xarray.DataArray
            - analogs
                - xarray.DataArray
            - events
                - xarray.Dataset
        - Right (context)
            ...

    Args:
        file_path: The path to the HDF5 file.
    """
    with netcdf.File(file_path, "r") as f:
        trial = _load_trial(f, file_path)

    return trial


def _load_trial(group: netcdf.File, file_path: Path) -> Trial:
    """Loads a trial from an HDF5 group.

    following structure is expected:
    - groupy (hdf5)
        - Left (context)
            - markers
                - xarray.DataArray
            - analogs
                - xarray.DataArray
            - events
                - xarray.Dataset
        - Right (context)
            ...
    Args:
        group: The group containing the trial data.
        file_path: The path to the HDF5 file.

    Returns:
        A new instance of the Trial class.
    """
    correct_file_format = False
    trial = Trial()
    for category in DataCategory:
        if category.value in group.groups.keys():
            with xr.load_dataarray(
                file_path, group=f"{group.name}/{category.value}"
            ) as data:
                trial.add_data(category, data)

            correct_file_format = True
    if "events" in group.groups.keys():
        with xr.load_dataset(
            file_path, group=f"{group.name}/events", engine="h5netcdf"
        ) as events:
            trial.events = events.to_dataframe()
        correct_file_format = True

    if not correct_file_format:
        raise ValueError(f"File {file_path} does not have the correct format.")

    return trial



import shutil
from pathlib import Path

import pandas as pd
import pytest

import gaitalytics.events as events
import gaitalytics.api as api
import gaitalytics.mapping as mapping
import gaitalytics.model as model


@pytest.fixture()
def out_path(request):
    out = Path('out/test_small_events.c3d')

    if out.exists():
        out.unlink()
    elif not out.parent.exists():
        out.parent.mkdir(parents=True)
    return out


@pytest.fixture()
def out_folder(request):
    out = Path('out/test_small')

    if out.exists():
        shutil.rmtree(out)
    elif not out.exists():
        out.mkdir(parents=True)
    return out


def test_load_config():
    config = api.load_config("./tests/pig_config.yaml")
    marker_name = config.get_marker_mapping(mapping.MappedMarkers.SACRUM)
    assert marker_name == "SACR"


def test_load_c3d_trial():
    config = api.load_config("./tests/pig_config.yaml")
    trial = api.load_c3d_trial("./tests/treadmill_events.c3d", config)
    assert len(trial.get_all_data().keys()) == 3
    assert trial.events is not None


def test_load_c3d_trial_no_events():
    config = api.load_config("./tests/pig_config.yaml")
    trial = api.load_c3d_trial("./tests/treadmill_no_events.c3d", config)
    assert trial.events is None


def test_detect_events():
    config = api.load_config("./tests/pig_config.yaml")
    trial = api.load_c3d_trial("./tests/treadmill_events.c3d", config)
    event_table = api.detect_events(trial, config, distance=1000)
    assert event_table is not None
    assert len(event_table) == 5


def test_detect_events_methode():
    config = api.load_config("./tests/pig_config.yaml")
    trial = model.Trial()
    with pytest.raises(TypeError):
        api.detect_events(trial, config, method= events.BaseEventDetection)


def test_check_events():
    config = api.load_config("./tests/pig_config.yaml")
    trial = api.load_c3d_trial("./tests/treadmill_events.c3d", config)
    api.check_events(trial.events)
    assert True


def test_check_events_methode():
    with pytest.raises(ValueError):
        api.check_events(pd.DataFrame(), method="foo")


def test_write_events(out_path):
    config = api.load_config("./tests/pig_config.yaml")
    trial = api.load_c3d_trial("./tests/treadmill_events.c3d", config)
    event_table = api.detect_events(trial, config, distance=1000)
    api.write_events_to_c3d("./tests/treadmill_no_events.c3d", event_table, out_path)
    assert True


def test_segment_trial():
    config = api.load_config("./tests/pig_config.yaml")
    trial = api.load_c3d_trial("./tests/treadmill_events.c3d", config)
    segm_trial = api.segment_trial(trial)
    assert len(segm_trial.get_all_cycles().keys()) == 2


def test_segment_trial_no_events():
    trial = model.Trial()
    with pytest.raises(ValueError):
        api.segment_trial(trial)


def test_segment_trial_method():
    trial = model.Trial()
    with pytest.raises(ValueError):
        api.segment_trial(trial, method="foo")


def test_time_normalize_trail():
    config = api.load_config("./tests/pig_config.yaml")
    trial = api.load_c3d_trial("./tests/treadmill_events.c3d", config)
    norm_trial = api.time_normalise_trial(trial)
    assert norm_trial is not None


def test_time_normalize_cycle_trial():
    config = api.load_config("./tests/pig_config.yaml")
    trial = api.load_c3d_trial("./tests/treadmill_events.c3d", config)
    trial_cycles = api.segment_trial(trial)
    norm_trial = api.time_normalise_trial(trial_cycles)
    assert norm_trial is not None


def test_time_normalize_trial_frames():
    config = api.load_config("./tests/pig_config.yaml")
    trial = api.load_c3d_trial("./tests/treadmill_events.c3d", config)
    trial_cycles = api.segment_trial(trial)
    norm_trial = api.time_normalise_trial(trial_cycles, n_frames=200)
    markers = norm_trial.get_cycle("Left", 0).get_data(model.DataCategory.MARKERS)
    assert markers.shape[2] == 200


def test_calculate_features():
    config = api.load_config("./tests/pig_config.yaml")
    trial = api.load_c3d_trial("./tests/treadmill_events.c3d", config)
    trial_cycles = api.segment_trial(trial)
    features = api.calculate_features(trial_cycles, config)
    assert features.shape == (2, 10, 2281)


def test_export_trial(out_folder):
    config = api.load_config("./tests/pig_config.yaml")
    trial = api.load_c3d_trial("./tests/treadmill_events.c3d", config)
    api.export_trial(trial, out_folder)
    assert (out_folder / "markers.nc").exists()
    assert (out_folder / "analogs.nc").exists()
    assert (out_folder / "analysis.nc").exists()
    assert (out_folder / "events.nc").exists()

def test_export_segmented_trial(out_folder):
    config = api.load_config("./tests/pig_config.yaml")
    trial = api.load_c3d_trial("./tests/treadmill_events.c3d", config)
    trial_cycles = api.segment_trial(trial)
    api.export_trial(trial_cycles, out_folder)
    assert (out_folder / "markers.nc").exists()
    assert (out_folder / "analogs.nc").exists()
    assert (out_folder / "analysis.nc").exists()



"""gaitalytics is a Python package for gait analysis.

It provides tools for processing and analysing gait data, including segmentation,
normalisation, and feature extraction.
The package is designed to be flexible and extensible, functionality to suit their needs

"""

from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("package-name")
except PackageNotFoundError:
    # package is not installed
    pass



"""This module contains classes for checking and detecting events in a trial."""

from abc import ABC, abstractmethod

import numpy as np
import pandas as pd
import scipy as sp
import xarray as xr

import gaitalytics.io as io
import gaitalytics.mapping as mapping
import gaitalytics.model as model
import gaitalytics.utils.mocap as mocap

FOOT_STRIKE = "Foot Strike"
FOOT_OFF = "Foot Off"


class _BaseEventChecker(ABC):
    """Abstract class for event checkers.

    This class provides a common interface for checking events in a trial,
    which makes them interchangeable.
    """

    @abstractmethod
    def check_events(self, events: pd.DataFrame) -> tuple[bool, list | None]:
        """Checks the events in the trial.

        Args:
            events: The events to be checked.

        Returns:
            bool: True if the events are correct, False otherwise.
            list | None: A list of incorrect time slices,
                or None if the events are correct.
        """
        raise NotImplementedError


class SequenceEventChecker(_BaseEventChecker):
    """A class for checking the sequence of events in a trial.

    This class provides a method to check the sequence of events in a trial.
    It checks the sequence of event labels and contexts.
    """

    _TIME_COLUMN = io._EventInputFileReader.COLUMN_TIME
    _LABEL_COLUMN = io._EventInputFileReader.COLUMN_LABEL
    _CONTEXT_COLUMN = io._EventInputFileReader.COLUMN_CONTEXT

    def check_events(self, events: pd.DataFrame) -> tuple[bool, list | None]:
        """Checks the sequence of events in the trial.

        In a normal gait cycle, the sequence of events is as follows:
        1. Foot Strike (right)
        2. Foot Off (left)
        3. Foot Strike (left)
        4. Foot Off (right)

        Args:
            events: The events to be checked.

        Returns:
            bool: True if the sequence is correct, False otherwise.
            list | None: A list time slice of incorrect sequence,
            or None if the sequence is correct.

        """
        if events is None:
            raise ValueError("Trial does not have events.")

        incorrect_times = []

        incorrect_labels = self._check_labels(events)
        incorrect_contexts = self._check_contexts(events)

        if incorrect_labels:
            incorrect_times.append(incorrect_labels)
        if incorrect_contexts:
            incorrect_times.append(incorrect_contexts)

        return not bool(incorrect_times), incorrect_times if incorrect_times else None

    def _check_labels(self, events: pd.DataFrame) -> list[tuple]:
        """Check alternating sequence of event labels.

        Expected sequence of event labels:
        1. Foot Strike
        2. Foot Off
        3. Foot Strike
        4. Foot Off

        Args:
            events: The events to be checked.

        Returns:
            A list of incorrect time slices.
        """
        incorrect_times = []
        last_label = None
        last_time = None
        for i, label in enumerate(events[self._LABEL_COLUMN]):
            time = events[self._TIME_COLUMN].iloc[i]
            if label == last_label:
                incorrect_times.append((last_time, time))

            last_time = time
            last_label = label
        return incorrect_times

    def _check_contexts(self, events: pd.DataFrame) -> list[tuple]:
        """Check sequence of contexts of events.

        Expected sequence of event contexts:
        1. Right
        2. Right
        3. Left
        4. Left

        Args:
            events: The events to be checked.

        Returns:
            A list of incorrect time slices.
        """
        incorrect_times = []
        # Check the occurrence of the context in windows of 3 events.
        for i in range(len(events) - 3):
            max_occurance = (
                events[self._CONTEXT_COLUMN].iloc[i : i + 3].value_counts().max()
            )

            # If the context occurs more than twice in the window, it is incorrect.
            if max_occurance > 2:
                incorrect_times.append(
                    (
                        events[self._TIME_COLUMN].iloc[i],
                        events[self._TIME_COLUMN].iloc[i + 3],
                    )
                )

        return incorrect_times


class BaseEventDetection(ABC):
    """Abstract class for event detectors.

    This class provides a common interface for detecting events in a trial,
    which makes them interchangeable.
    """

    def __init__(self, configs: mapping.MappingConfigs):
        """Initializes a new instance of the BaseEventDetection class.

        Args:
            configs: The mapping configurations.
        """
        self._configs = configs

    @abstractmethod
    def detect_events(self, trial: model.Trial) -> pd.DataFrame:
        """Detects the events in the trial.

        Args:
            trial: The trial for which to detect the events.

        Returns:
            pd.DataFrame: A DataFrame containing the detected events.
        """
        raise NotImplementedError


class MarkerEventDetection(BaseEventDetection):
    """A class for detecting events using marker data.

    This class provides a method to detect events using marker data in a trial.
    The algorithm is based on the paper by Zeni et al. (2008).
    """

    _TIME_COLUMN = io._EventInputFileReader.COLUMN_TIME
    _LABEL_COLUMN = io._EventInputFileReader.COLUMN_LABEL
    _CONTEXT_COLUMN = io._EventInputFileReader.COLUMN_CONTEXT
    _ICON_COLUMN = io._EventInputFileReader.COLUMN_ICON

    def __init__(self, configs: mapping.MappingConfigs, **kwargs):
        """Initializes a new instance of the MarkerEventDetection class.

        Args:
            configs: The mapping configurations.
            **kwargs:
                - height: The height of peaks. Default = None
                - threshold: The threshold of peaks. Default = None
                - distance: The min distance in frames between events. Default = None
                - rel_height: The relative height of peak. Default = 0.5
        """

        self._height = kwargs.get("height", None)
        self._threshold = kwargs.get("threshold", None)
        self._distance = kwargs.get("distance", None)
        self._rel_height = kwargs.get("rel_height", 0.5)
        super().__init__(configs)

    def detect_events(self, trial: model.Trial) -> pd.DataFrame:
        """Detects the events in the trial using marker data.

        Args:
            trial: The trial for which to detect the events.

        Returns:
            pd.DataFrame: A DataFrame containing the detected events.
        """
        scarum = mocap.get_sacrum_marker(trial, self._configs)
        l_heel = mocap.get_marker_data(
            trial, self._configs, mapping.MappedMarkers.L_HEEL
        )
        r_heel = mocap.get_marker_data(
            trial, self._configs, mapping.MappedMarkers.R_HEEL
        )
        l_toe = mocap.get_marker_data(trial, self._configs, mapping.MappedMarkers.L_TOE)
        r_toe = mocap.get_marker_data(trial, self._configs, mapping.MappedMarkers.R_TOE)
        sacrum = mocap.get_sacrum_marker(trial, self._configs)
        l_heel, l_toe, r_heel, r_toe = self._rotate_markers(
            l_heel, l_toe, r_heel, r_toe, sacrum, scarum, trial
        )

        l_hs_times = self._detect_events(scarum, l_heel, False)
        r_hs_times = self._detect_events(scarum, r_heel, False)
        l_to_times = self._detect_events(scarum, l_toe, True)
        r_to_times = self._detect_events(scarum, r_toe, True)

        l_hs_events = self._create_data_frame(l_hs_times, "Left", FOOT_STRIKE)
        r_hs_events = self._create_data_frame(r_hs_times, "Right", FOOT_STRIKE)
        l_to_events = self._create_data_frame(l_to_times, "Left", FOOT_OFF)
        r_to_events = self._create_data_frame(r_to_times, "Right", FOOT_OFF)

        events = pd.concat([l_hs_events, r_hs_events, l_to_events, r_to_events])
        events = events.sort_values(by=self._TIME_COLUMN, ascending=True).reset_index(
            drop=True
        )
        return events

    def _rotate_markers(self, l_heel, l_toe, r_heel, r_toe, sacrum, scarum, trial):
        l_ant_hip = mocap.get_marker_data(
            trial, self._configs, mapping.MappedMarkers.L_ANT_HIP
        )
        r_ant_hip = mocap.get_marker_data(
            trial, self._configs, mapping.MappedMarkers.R_ANT_HIP
        )
        ant_hip = (l_ant_hip + r_ant_hip) / 2
        progress_axis = ant_hip - sacrum
        x_axis = xr.DataArray(
            [1, 0, 0], dims=["axis"], coords={"axis": ["x", "y", "z"]}
        )
        angles = self._calculate_angle(progress_axis, x_axis)
        l_heel = self._rotate_point(l_heel, scarum, angles)
        r_heel = self._rotate_point(r_heel, scarum, angles)
        l_toe = self._rotate_point(l_toe, scarum, angles)
        r_toe = self._rotate_point(r_toe, scarum, angles)
        ant_hip = self._rotate_point(ant_hip, scarum, angles)

        scale = self._get_flip_scale(ant_hip - sacrum)
        l_heel = (l_heel.T * scale).T
        r_heel = (r_heel.T * scale).T
        l_toe = (l_toe.T * scale).T
        r_toe = (r_toe.T * scale).T

        return l_heel, l_toe, r_heel, r_toe

    @staticmethod
    def _calculate_angle(progress: xr.DataArray, axis: xr.DataArray) -> float:
        """Calculate the angle between two vectors.

        Args:
            progress: The first vector.
            axis: The second vector.

        Returns:
            float: The angle between the two vectors.
        """
        progress = progress.drop_sel(axis="z")
        axis = axis.drop_sel(axis="z")
        theta = np.arccos(
            progress.dot(axis, dim="axis")
            / (progress.meca.norm(dim="axis") * axis.meca.norm(dim="axis"))
        )
        return theta.values

    @staticmethod
    def _rotate_point(
        point: xr.DataArray, fix_point: xr.DataArray, angle: float
    ) -> xr.DataArray:
        """Rotate a point around a fixed point.

        Args:
            point: The point to rotate.
            fix_point: The fixed point.
            angle: The angle to rotate by.

        Returns:
            xr.DataArray: The rotated point.
        """
        fix = fix_point.drop_sel(axis="z")
        rel_point = point.drop_sel(axis="z")
        rot = np.array(
            [[np.cos(angle), -np.sin(angle)], [np.sin(angle), np.cos(angle)]]
        )
        rel_fix = fix.to_numpy()
        np_rel_point = rel_point.to_numpy()
        two_d_point = np.empty(np_rel_point.shape)
        for i in range(np_rel_point.shape[1]):
            two_d_point[:, i] = np_rel_point[:, i] - rel_fix[:, i]
            two_d_point[:, i] = two_d_point[:, i] @ rot[:, :, i].T
            two_d_point[:, i] = two_d_point[:, i] + rel_fix[:, i]
        point = point.copy(deep=True)
        point.loc["x"] = two_d_point[0]
        point.loc["y"] = two_d_point[1]
        return point

    @staticmethod
    def _get_flip_scale(rot_progress: xr.DataArray) -> list:
        if (rot_progress.loc["x"] < 0).sum() > rot_progress.loc["x"].shape[0] / 2:
            return [-1, 1, 1]
        else:
            return [1, 1, 1]

    def _detect_events(
        self, sacrum: xr.DataArray, point: xr.DataArray, toe_off: bool
    ) -> np.ndarray:
        """Detects the events in the trial using projected points.
        Args:
            sacrum: The projected sacrum point.
            point: The projected point to detect the events for.
            toe_off: True if the event is toe off, False otherwise.

        Returns:
            np.ndarray: The detected event times.
        """
        if toe_off:
            distance = sacrum - point
        else:
            distance = point - sacrum
        distance = distance.sel(axis="x").meca.normalize().meca.center()
        index, heights = sp.signal.find_peaks(
            distance.to_numpy(),
            height=self._height,
            threshold=self._threshold,
            distance=self._distance,
            rel_height=self._rel_height,
        )

        # take smaller peaks if toe_off, larger peaks if heel_strike

        times = point[:, index].coords["time"].values

        return times

    def _create_data_frame(
        self, times: np.ndarray, context: str, label: str
    ) -> pd.DataFrame:
        """Creates a DataFrame from the detected events.


        Args:
            times: The detected event times.
            context: The context of the detected events.
            label: The label of the detected events.

        Returns:
            pd.DataFrame: A DataFrame containing the detected events.
        """
        contexts = [context] * len(times)
        labels = [label] * len(times)
        icons = [1 if label == FOOT_STRIKE else 2] * len(times)

        table = {
            self._TIME_COLUMN: times,
            self._LABEL_COLUMN: labels,
            self._CONTEXT_COLUMN: contexts,
            self._ICON_COLUMN: icons,
        }
        events = pd.DataFrame.from_dict(table)
        return events



import xarray as xr

import gaitalytics.mapping as mapping
import gaitalytics.model as model


def get_marker_data(
    trial: model.Trial, config: mapping.MappingConfigs, marker: mapping.MappedMarkers
) -> xr.DataArray:
    """Get the marker data for a trial.

    Args:
        trial: The trial for which to get the marker data.
        config: The mapping configurations.
        marker: The marker to get the data for.

    Returns:
        An xarray DataArray containing the marker data.
    """
    marker_label = config.get_marker_mapping(marker)
    markers = trial.get_data(model.DataCategory.MARKERS)
    marker_data = markers.sel(channel=marker_label)
    return marker_data


def get_sacrum_marker(
    trial: model.Trial, config: mapping.MappingConfigs
) -> xr.DataArray:
    """Get the sacrum marker data for a trial.

    Try to get the sacrum marker data from the trial.
    If the sacrum marker not found calculate from posterior hip markers

    Args:
        trial: The trial for which to get the marker data.
        config: The mapping configurations.

    Returns:
        An xarray DataArray containing the sacrum marker data.
    """
    try:
        return get_marker_data(trial, config, mapping.MappedMarkers.SACRUM)
    except KeyError:
        l_marker = get_marker_data(trial, config, mapping.MappedMarkers.L_POST_HIP)
        r_marker = get_marker_data(trial, config, mapping.MappedMarkers.R_POST_HIP)
        return (l_marker + r_marker) / 2


def get_progression_vector(
    trial: model.Trial, config: mapping.MappingConfigs
) -> xr.DataArray:
    """Calculate the progression vector for a trial.

    The progression vector is the vector from the sacrum to the anterior hip marker.

    Args:
        trial: The trial for which to calculate the progression vector.
        config: The mapping configurations.

    Returns:
        An xarray DataArray containing the calculated progression vector.
    """
    sacrum_marker = get_sacrum_marker(trial, config)
    r_ant_hip = get_marker_data(trial, config, mapping.MappedMarkers.R_ANT_HIP)
    l_ant_hip = get_marker_data(trial, config, mapping.MappedMarkers.L_ANT_HIP)
    ant_marker = (r_ant_hip + l_ant_hip) / 2

    return (ant_marker - sacrum_marker).mean(dim="time")
